# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str

from .client import client as cls_client
from .cluster import cluster as cls_cluster
from .ctx import ctx as cls_ctx
from .ip import ip as cls_ip
from .log import log as cls_log
from .nginx import nginx as cls_nginx
from .node import node as cls_node
from .request import request as cls_request
from .response import response as cls_response
from .router import router as cls_router
from .service import service as cls_service
from .vault import vault as cls_vault

class kong():

    client = cls_client
    cluster = cls_cluster
    ctx = cls_ctx
    ip = cls_ip
    log = cls_log
    nginx = cls_nginx
    node = cls_node
    request = cls_request
    response = cls_response
    router = cls_router
    service = cls_service
    vault = cls_vault

    pass


# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

import os
import sys
import argparse
import msgpack
import json

from .server import PluginServer
from .listener import UnixStreamServer, DEFAULT_SOCKET_NAME
from .logger import Logger

from .const import __version__, PY3K

def parse(dedicated=False):
    parser = argparse.ArgumentParser(description='Kong Python Plugin Server.')
    parser.add_argument('-p', '--kong-prefix', '-kong-prefix',
                        dest='prefix', metavar='prefix', type=str,
                        default="/usr/local/kong/",
                        help='unix domain socket path to listen (default: %(default)s)')
    parser.add_argument('-v', '--verbose', action='count', default=1,
                        help='turn on verbose logging (default: %(default)s)')
    parser.add_argument('--version', '-version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument('--socket-name', type=str, dest='socket_name', default=DEFAULT_SOCKET_NAME,
                        help='socket name to listen on (default: %(default)s)')
    parser.add_argument('--listen-queue-size', type=int, dest='listen_queue_size', default=4096,
                        help='socket listen queue size (default: %(default)s)')
    parser.add_argument('--no-lua-style', action='store_true', dest='no_lua_style', default=False,
                        help='turn off Lua-style "data, err" return values for PDK functions '
                             'and throw exception instead (default: %(default)s)')
    mxg = parser.add_mutually_exclusive_group()
    mxg.add_argument('-m', '--multiprocessing', dest='multiprocessing', action="store_true",
                     help='enable multiprocessing (default: %(default)s)')
    mxg.add_argument('-g', '--gevent', dest='gevent', action="store_true",
                     help='enable gevent (default: %(default)s)')

    if not dedicated:
        parser.add_argument('-d', '--plugins-directory', '-plugins-directory',
                            dest='directory', metavar='directory', type=str, required=True,
                            help='plugins directory')
        parser.add_argument('--dump-plugin-info', '-dump-plugin-info', dest='dump_info', metavar='name', type=str,
                            help='dump specific plugin info into stdout')
        parser.add_argument('--dump-all-plugins', '-dump-all-plugins', dest='dump_all_info', action="store_true",
                            help='dump all plugins info into stdout')
    else:
        parser.add_argument('--dump', '-dump', dest='dump', action="store_true",
                            help='dump current plugin info into stdout')

    args = parser.parse_args()

    if ((not dedicated and not args.dump_info and not args.dump_all_info)
            or (dedicated and not args.dump)) and not os.path.exists(args.prefix):
        raise OSError("path %s doesn't exist, can't create unix socket file" % args.prefix)

    return args

def display_lua_style_notice(lua_style, ps):
    if not lua_style:
        ps.logger.debug("python-style return values are enabled, "
                        "errors are thrown as exception")
    else:
        ps.logger.warn("lua-style return values are used, "
                       "this will be deprecated in the future; instead of returning "
                       "(data, err) tuple, only data will be returned and err will be "
                       "thrown as PDKException; please adjust your plugin to use the "
                       "new python-style PDK API.")

def start_server():
    args = parse()

    prefix = args.prefix

    ps = PluginServer(loglevel=Logger.WARNING - args.verbose,
                      plugin_dir=args.directory,
                      use_multiprocess=args.multiprocessing,
                      use_gevent=args.gevent,
                      lua_style=not args.no_lua_style)
    ss = UnixStreamServer(ps, prefix,
                          sock_name=args.socket_name,
                          use_gevent=args.gevent,
                          listen_queue_size=args.listen_queue_size)
    if args.dump_info:
        ret, err = ps.get_plugin_info(args.dump_info)
        if err:
            raise Exception("error dump info: " + err)
        if PY3K:
            sys.stdout.buffer.write(msgpack.packb(ret))
        else:
            sys.stdout.write(msgpack.packb(ret))
        sys.exit(0)
    elif args.dump_all_info:
        ret = []
        for p in ps.plugins:
            inf = ps.get_plugin_info(p)
            ret.append(inf)
        sys.stdout.write(json.dumps(ret))
        sys.exit(0)

    display_lua_style_notice(not args.no_lua_style, ps)

    try:
        ss.serve_forever()
    except KeyboardInterrupt:
        ps.logger.info("polite exit requested, terminating...")

    ps.cleanup()

def start_dedicated_server(name, plugin, _version=None, _priority=0, _schema=[]):
    from .module import Module

    args = parse(dedicated=True)

    ps = PluginServer(loglevel=Logger.WARNING - args.verbose,
                      use_multiprocess=args.multiprocessing,
                      use_gevent=args.gevent,
                      name="%s version %s" % (name, _version or 'unknown'),
                      lua_style=not args.no_lua_style)
    socket_name = args.socket_name
    if socket_name == DEFAULT_SOCKET_NAME:
        socket_name = "%s.sock" % name
    ss = UnixStreamServer(ps, args.prefix,
                          sock_name=socket_name,
                          use_gevent=args.gevent,
                          listen_queue_size=args.listen_queue_size)

    class mod(object):
        Plugin = plugin
        Schema = _schema
        version = _version
        priority = _priority

    mod = Module(name, module=mod)
    ps.plugins[name] = mod

    if args.dump:
        ret = ps.get_plugin_info(name)
        # note a list is returned
        sys.stdout.write(json.dumps([ret]))
        sys.exit(0)

    display_lua_style_notice(not args.no_lua_style, ps)

    try:
        ss.serve_forever()
    except KeyboardInterrupt:
        ps.logger.info("polite exit requested, terminating...")

    ps.cleanup()



# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/request.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class request():


    @staticmethod
    def get_body(mimetype: Optional[str], max_args: Optional[number]) -> Tuple[table, str, str]:
        """

            Returns the request data as a key/value table.
            A high-level convenience function.
            The body is parsed with the most appropriate format:
            * If `mimetype` is specified, it decodes the body with the requested
              content type (if supported). This takes precedence over any content type
              present in the request.
              The optional argument `mimetype` can be one of the following strings:
                * `application/x-www-form-urlencoded`
                * `application/json`
                * `multipart/form-data`
            Whether `mimetype` is specified or a request content type is otherwise
            present in the request, each content type behaves as follows:
            * If the request content type is `application/x-www-form-urlencoded`:
              * Returns the body as form-encoded.
            * If the request content type is `multipart/form-data`:
              * Decodes the body as multipart form data
                (same as `multipart(kong.request.get_raw_body(),
                kong.request.get_header("Content-Type")):get_all()` ).
            * If the request content type is `application/json`:
              * Decodes the body as JSON
                (same as `json.decode(kong.request.get_raw_body())`).
              * JSON types are converted to matching Lua types.
            * If the request contains none of the above and the `mimetype` argument is
              not set, returns `nil` and an error message indicating the
              body could not be parsed.
            The optional argument `max_args` can be used to set a limit on the number
            of form arguments parsed for `application/x-www-form-urlencoded` payloads.
            The third return value is string containing the mimetype used to parsed
            the body (as per the `mimetype` argument), allowing the caller to identify
            what MIME type the body was parsed as.

        Phases:
            rewrite, access, response, admin_api

        Example:
            body, err, mimetype = kong.request.get_body()

            body.name # "John Doe"

            body.age  # "42"

        :parameter mimetype: The MIME type.
        :type mimetype: str
        :parameter max_args: Sets a limit on the maximum number of parsed
            arguments.
        :type max_args: number

        :return: A table representation of the body.

        :rtype: table
        :return: An error message.

        :rtype: str
        :return: mimetype The MIME type used.

        :rtype: str
        """
        pass

    @staticmethod
    def get_forwarded_host() -> str:
        """

            Returns the host component of the request's URL or the value of the "host"
            header. Unlike `kong.request.get_host()`, this function also considers
            `X-Forwarded-Host` if it comes from a trusted source. The returned value
            is normalized to lowercase.
            Whether this function considers `X-Forwarded-Host` or not depends on
            several Kong configuration parameters:
            * [trusted\_ips](https://docs.konghq.com/gateway/latest/reference/configuration/#trusted_ips)
            * [real\_ip\_header](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_header)
            * [real\_ip\_recursive](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_recursive)
            **Note**: Kong does not offer support for the Forwarded HTTP Extension
            (RFC 7239) since it is not supported by ngx_http_realip_module.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            kong.request.get_forwarded_host() # "example.com"

        :return: The forwarded host.

        :rtype: str
        """
        pass

    @staticmethod
    def get_forwarded_path() -> str:
        """

            Returns the path component of the request's URL, but also considers
            `X-Forwarded-Path` if it comes from a trusted source. The value
            is returned as a Lua string.
            Whether this function considers `X-Forwarded-Path` or not depends on
            several Kong configuration parameters:
            * [trusted\_ips](https://docs.konghq.com/gateway/latest/reference/configuration/#trusted_ips)
            * [real\_ip\_header](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_header)
            * [real\_ip\_recursive](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_recursive)
            **Note**: Kong does not do any normalization on the request path.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            kong.request.get_forwarded_path() # /path

        :return: The forwarded path.

        :rtype: str
        """
        pass

    @staticmethod
    def get_forwarded_port() -> number:
        """

            Returns the port component of the request's URL, but also considers
            `X-Forwarded-Host` if it comes from a trusted source. The value
            is returned as a Lua number.
            Whether this function considers `X-Forwarded-Proto` or not depends on
            several Kong configuration parameters:
            * [trusted\_ips](https://docs.konghq.com/gateway/latest/reference/configuration/#trusted_ips)
            * [real\_ip\_header](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_header)
            * [real\_ip\_recursive](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_recursive)
            **Note**: Kong does not offer support for the Forwarded HTTP Extension
            (RFC 7239) since it is not supported by ngx_http_realip_module.
            When running Kong behind the L4 port mapping (or forwarding), you can also
            configure:
            * [port\_maps](https://docs.konghq.com/gateway/latest/reference/configuration/#port_maps)
            The `port_maps` configuration parameter enables this function to return the
            port to which the port Kong is listening to is mapped to (in case they differ).

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            kong.request.get_forwarded_port() # 1234

        :return: The forwarded port.

        :rtype: number
        """
        pass

    @staticmethod
    def get_forwarded_prefix() -> str:
        """

            Returns the prefix path component of the request's URL that Kong stripped
            before proxying to upstream. It also checks if `X-Forwarded-Prefix` comes
            from a trusted source, and uses it as-is when given. The value is returned
            as a Lua string.
            If a trusted `X-Forwarded-Prefix` is not passed, this function must be
            called after Kong has run its router (`access` phase),
            as the Kong router may strip the prefix of the request path. That stripped
            path becomes the return value of this function, unless there is already
            a trusted `X-Forwarded-Prefix` header in the request.
            Whether this function considers `X-Forwarded-Prefix` or not depends on
            several Kong configuration parameters:
            * [trusted\_ips](https://docs.konghq.com/gateway/latest/reference/configuration/#trusted_ips)
            * [real\_ip\_header](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_header)
            * [real\_ip\_recursive](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_recursive)
            **Note**: Kong does not do any normalization on the request path prefix.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            kong.request.get_forwarded_prefix() # /prefix

        :return: The forwarded path prefix or `nil` if the prefix was
            not stripped.

        :rtype: str
        """
        pass

    @staticmethod
    def get_forwarded_scheme() -> str:
        """

            Returns the scheme component of the request's URL, but also considers
            `X-Forwarded-Proto` if it comes from a trusted source. The returned
            value is normalized to lowercase.
            Whether this function considers `X-Forwarded-Proto` or not depends on
            several Kong configuration parameters:
            * [trusted\_ips](https://docs.konghq.com/gateway/latest/reference/configuration/#trusted_ips)
            * [real\_ip\_header](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_header)
            * [real\_ip\_recursive](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_recursive)
            **Note**: Kong does not offer support for the Forwarded HTTP Extension
            (RFC 7239) since it is not supported by ngx_http_realip_module.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            kong.request.get_forwarded_scheme() # "https"

        :return: The forwarded scheme.

        :rtype: str
        """
        pass

    @staticmethod
    def get_header(name: str) -> str:
        """

            Returns the value of the specified request header.
            The returned value is either a `string`, or can be `nil` if a header with
            `name` was not found in the request. If a header with the same name is
            present multiple times in the request, this function returns the value
            of the first occurrence of this header.
            Header names in are case-insensitive and are normalized to lowercase, and
            dashes (`-`) can be written as underscores (`_`); that is, the header
            `X-Custom-Header` can also be retrieved as `x_custom_header`.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            # Given a request with the following headers:

            # Host: foo.com

            # X-Custom-Header: bla

            # X-Another: foo bar

            # X-Another: baz

            kong.request.get_header("Host")            # "foo.com"

            kong.request.get_header("x-custom-header") # "bla"

            kong.request.get_header("X-Another")       # "foo bar"

        :parameter name: the name of the header to be returned
        :type name: str

        :return: the value of the header or nil if not present

        :rtype: str
        """
        pass

    @staticmethod
    def get_headers(max_headers: Optional[number]) -> table:
        """

            Returns a Lua table holding the request headers. Keys are header names.
            Values are either a string with the header value, or an array of strings
            if a header was sent multiple times. Header names in this table are
            case-insensitive and are normalized to lowercase, and dashes (`-`) can be
            written as underscores (`_`); that is, the header `X-Custom-Header` can
            also be retrieved as `x_custom_header`.
            By default, this function returns up to **100** headers. The optional
            `max_headers` argument can be specified to customize this limit, but must
            be greater than **1** and not greater than **1000**.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            # Given a request with the following headers:

            # Host: foo.com

            # X-Custom-Header: bla

            # X-Another: foo bar

            # X-Another: baz

            headers = kong.request.get_headers()

            headers.host            # "foo.com"

            headers.x_custom_header # "bla"

            headers.x_another[1]    # "foo bar"

            headers["X-Another"][2] # "baz"

        :parameter max_headers: Sets a limit on the maximum number of
            parsed headers.
        :type max_headers: number

        :return: The request headers in table form.

        :rtype: table
        """
        pass

    @staticmethod
    def get_host() -> str:
        """

            Returns the host component of the request's URL, or the value of the
            "Host" header. The returned value is normalized to lowercase form.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            # Given a request to https://example.com:1234/v1/movies

            kong.request.get_host() # "example.com"

        :return: The hostname.

        :rtype: str
        """
        pass

    @staticmethod
    def get_http_version() -> number:
        """

            Returns the HTTP version used by the client in the request as a Lua
            number, returning values such as `1`, `1.1`, `2.0`, or `nil` for
            unrecognized values.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            kong.request.get_http_version() # 1.1

        :return: The HTTP version as a Lua number.

        :rtype: number
        """
        pass

    @staticmethod
    def get_method() -> str:
        """

            Returns the HTTP method of the request. The value is normalized to
            uppercase.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            kong.request.get_method() # "GET"

        :return: The request method.

        :rtype: str
        """
        pass

    @staticmethod
    def get_path() -> str:
        """

            Returns the path component of the request's URL. It is not normalized in
            any way and does not include the query string.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            # Given a request to https://example.com:1234/v1/movies?movie=foo

            kong.request.get_path() # "/v1/movies"

        :return: The path.

        :rtype: str
        """
        pass

    @staticmethod
    def get_path_with_query() -> str:
        """

            Returns the path, including the query string if any. No
            transformations or normalizations are done.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            # Given a request to https://example.com:1234/v1/movies?movie=foo

            kong.request.get_path_with_query() # "/v1/movies?movie=foo"

        :return: The path with the query string.

        :rtype: str
        """
        pass

    @staticmethod
    def get_port() -> number:
        """

            Returns the port component of the request's URL. The value is returned
            as a Lua number.

        Phases:
            certificate, rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            # Given a request to https://example.com:1234/v1/movies

            kong.request.get_port() # 1234

        :return: The port.

        :rtype: number
        """
        pass

    @staticmethod
    def get_query(max_args: Optional[number]) -> table:
        """

            Returns the table of query arguments obtained from the query string. Keys
            are query argument names. Values are either a string with the argument
            value, a boolean `true` if an argument was not given a value, or an array
            if an argument was given in the query string multiple times. Keys and
            values are unescaped according to URL-encoded escaping rules.
            Note that a query string `?foo&bar` translates to two boolean `true`
            arguments, and `?foo=&bar=` translates to two string arguments containing
            empty strings.
            By default, this function returns up to **100** arguments. The optional
            `max_args` argument can be specified to customize this limit, but must be
            greater than **1** and not greater than **1000**.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            # Given a request GET /test?foo=hello%20world&bar=baz&zzz&blo=&bar=bla&bar

            for k, v in pairs(kong.request.get_query()) do

            kong.log.inspect(k, v)# Will print

            # "foo" "hello world"

            # "bar" {"baz", "bla", true}

            # "zzz" true

            # "blo" ""

        :parameter max_args: Sets a limit on the maximum number of parsed
            arguments.
        :type max_args: number

        :return: A table representation of the query string.

        :rtype: table
        """
        pass

    @staticmethod
    def get_query_arg() -> Any:
        """

            Returns the value of the specified argument, obtained from the query
            arguments of the current request.
            The returned value is either a `string`, a boolean `true` if an
            argument was not given a value, or `nil` if no argument with `name` was
            found.
            If an argument with the same name is present multiple times in the
            query string, this function returns the value of the first occurrence.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            # Given a request GET /test?foo=hello%20world&bar=baz&zzz&blo=&bar=bla&bar

            kong.request.get_query_arg("foo") # "hello world"

            kong.request.get_query_arg("bar") # "baz"

            kong.request.get_query_arg("zzz") # true

            kong.request.get_query_arg("blo") # ""

        :return: The value of the argument.

        :rtype: Any
        """
        pass

    @staticmethod
    def get_raw_body() -> str:
        """

            Returns the plain request body.
            If the body has no size (empty), this function returns an empty string.
            If the size of the body is greater than the Nginx buffer size (set by
            `client_body_buffer_size`), this function fails and returns an error
            message explaining this limitation.

        Phases:
            rewrite, access, response, admin_api

        Example:
            # Given a body with payload "Hello, Earth!":

            kong.request.get_raw_body():gsub("Earth", "Mars") # "Hello, Mars!"

        :return: The plain request body.

        :rtype: str
        """
        pass

    @staticmethod
    def get_raw_query() -> str:
        """

            Returns the query component of the request's URL. It is not normalized in
            any way (not even URL-decoding of special characters) and does not
            include the leading `?` character.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            # Given a request to https://example.com/foo?msg=hello%20world&bla=&bar

            kong.request.get_raw_query() # "msg=hello%20world&bla=&bar"

        :return: The query component of the request's URL.

        :rtype: str
        """
        pass

    @staticmethod
    def get_scheme() -> str:
        """

            Returns the scheme component of the request's URL. The returned value is
            normalized to lowercase form.

        Phases:
            rewrite, access, header_filter, response, body_filter, log, admin_api

        Example:
            # Given a request to https://example.com:1234/v1/movies

            kong.request.get_scheme() # "https"

        :return: A string like `"http"` or `"https"`.

        :rtype: str
        """
        pass

    pass


# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/log.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class log():


    @staticmethod
    def alert(*args: Any) -> None:
        """

            Similar to `kong.log()`, but the produced log has the severity given by
            `<level>`, instead of `notice`. The supported levels are:
            * `kong.log.alert()`
            * `kong.log.crit()`
            * `kong.log.err()`
            * `kong.log.warn()`
            * `kong.log.notice()`
            * `kong.log.info()`
            * `kong.log.debug()`
            Logs have the same format as that of `kong.log()`. For
            example, the following call:
            ``` lua
             kong.log.err("hello ", "world")
            ```
            would, within the core, produce a log line similar to:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```
            If invoked from within a plugin (for example, `key-auth`) it would include the
            namespace prefix:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 [key-auth] hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```

        Phases:
            init_worker, certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            kong.log.warn("something require attention")

            kong.log.err("something failed: ", err)

            kong.log.alert("something requires immediate action")

        :parameter *args: All params will be concatenated and stringified before being sent to the log.
        :type *args: Any

        :return: Throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def crit(*args: Any) -> None:
        """

            Similar to `kong.log()`, but the produced log has the severity given by
            `<level>`, instead of `notice`. The supported levels are:
            * `kong.log.alert()`
            * `kong.log.crit()`
            * `kong.log.err()`
            * `kong.log.warn()`
            * `kong.log.notice()`
            * `kong.log.info()`
            * `kong.log.debug()`
            Logs have the same format as that of `kong.log()`. For
            example, the following call:
            ``` lua
             kong.log.err("hello ", "world")
            ```
            would, within the core, produce a log line similar to:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```
            If invoked from within a plugin (for example, `key-auth`) it would include the
            namespace prefix:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 [key-auth] hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```

        Phases:
            init_worker, certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            kong.log.warn("something require attention")

            kong.log.err("something failed: ", err)

            kong.log.alert("something requires immediate action")

        :parameter *args: All params will be concatenated and stringified before being sent to the log.
        :type *args: Any

        :return: Throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def debug(*args: Any) -> None:
        """

            Similar to `kong.log()`, but the produced log has the severity given by
            `<level>`, instead of `notice`. The supported levels are:
            * `kong.log.alert()`
            * `kong.log.crit()`
            * `kong.log.err()`
            * `kong.log.warn()`
            * `kong.log.notice()`
            * `kong.log.info()`
            * `kong.log.debug()`
            Logs have the same format as that of `kong.log()`. For
            example, the following call:
            ``` lua
             kong.log.err("hello ", "world")
            ```
            would, within the core, produce a log line similar to:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```
            If invoked from within a plugin (for example, `key-auth`) it would include the
            namespace prefix:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 [key-auth] hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```

        Phases:
            init_worker, certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            kong.log.warn("something require attention")

            kong.log.err("something failed: ", err)

            kong.log.alert("something requires immediate action")

        :parameter *args: All params will be concatenated and stringified before being sent to the log.
        :type *args: Any

        :return: Throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def deprecation(*args: Any) -> None:
        """

            Arguments given to this function can be of any type, but table arguments
            are converted to strings via `tostring` (thus potentially calling a
            table's `__tostring` metamethod if set). When the last argument is a table,
            it is considered as a deprecation metadata. The table can include the
            following properties:
            ``` lua
            {
              after = "2.5.0",   -- deprecated after Kong version 2.5.0 (defaults to `nil`)
              removal = "3.0.0", -- about to be removed with Kong version 3.0.0 (defaults to `nil`)
              trace = true,      -- writes stack trace along with the deprecation message (defaults to `nil`)
            }
            ```
            For example, the following call:
            ``` lua
            kong.log.deprecation("hello ", "world")
            ```
            would, within the core, produce a log line similar to:
            ``` plain
            2017/07/09 19:36:25 [warn] 25932#0: *1 [kong] some_file.lua:54 hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```
            If invoked from within a plugin (for example, `key-auth`) it would include the
            namespace prefix:
            ``` plain
            2017/07/09 19:36:25 [warn] 25932#0: *1 [kong] some_file.lua:54 [key-auth] hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```
            And with metatable, the following call:
            ``` lua
            kong.log.deprecation("hello ", "world", { after = "2.5.0", removal = "3.0.0" })
            ```
            would, within the core, produce a log line similar to:
            ``` plain
            2017/07/09 19:36:25 [warn] 25932#0: *1 [kong] some_file.lua:54 hello world (deprecated after 2.5.0, scheduled for removal in 3.0.0), client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```

        Phases:
            init_worker, certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            kong.log.deprecation("hello ", "world")

            kong.log.deprecation("hello ", "world", { after = "2.5.0" })

            kong.log.deprecation("hello ", "world", { removal = "3.0.0" })

            kong.log.deprecation("hello ", "world", { after = "2.5.0", removal = "3.0.0" })

            kong.log.deprecation("hello ", "world", { trace = true })

        :parameter *args: all params will be concatenated and stringified before being sent to the log
            (if the last param is a table, it is considered as a deprecation metadata)
        :type *args: Any

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def err(*args: Any) -> None:
        """

            Similar to `kong.log()`, but the produced log has the severity given by
            `<level>`, instead of `notice`. The supported levels are:
            * `kong.log.alert()`
            * `kong.log.crit()`
            * `kong.log.err()`
            * `kong.log.warn()`
            * `kong.log.notice()`
            * `kong.log.info()`
            * `kong.log.debug()`
            Logs have the same format as that of `kong.log()`. For
            example, the following call:
            ``` lua
             kong.log.err("hello ", "world")
            ```
            would, within the core, produce a log line similar to:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```
            If invoked from within a plugin (for example, `key-auth`) it would include the
            namespace prefix:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 [key-auth] hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```

        Phases:
            init_worker, certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            kong.log.warn("something require attention")

            kong.log.err("something failed: ", err)

            kong.log.alert("something requires immediate action")

        :parameter *args: All params will be concatenated and stringified before being sent to the log.
        :type *args: Any

        :return: Throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def info(*args: Any) -> None:
        """

            Similar to `kong.log()`, but the produced log has the severity given by
            `<level>`, instead of `notice`. The supported levels are:
            * `kong.log.alert()`
            * `kong.log.crit()`
            * `kong.log.err()`
            * `kong.log.warn()`
            * `kong.log.notice()`
            * `kong.log.info()`
            * `kong.log.debug()`
            Logs have the same format as that of `kong.log()`. For
            example, the following call:
            ``` lua
             kong.log.err("hello ", "world")
            ```
            would, within the core, produce a log line similar to:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```
            If invoked from within a plugin (for example, `key-auth`) it would include the
            namespace prefix:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 [key-auth] hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```

        Phases:
            init_worker, certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            kong.log.warn("something require attention")

            kong.log.err("something failed: ", err)

            kong.log.alert("something requires immediate action")

        :parameter *args: All params will be concatenated and stringified before being sent to the log.
        :type *args: Any

        :return: Throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def notice(*args: Any) -> None:
        """

            Similar to `kong.log()`, but the produced log has the severity given by
            `<level>`, instead of `notice`. The supported levels are:
            * `kong.log.alert()`
            * `kong.log.crit()`
            * `kong.log.err()`
            * `kong.log.warn()`
            * `kong.log.notice()`
            * `kong.log.info()`
            * `kong.log.debug()`
            Logs have the same format as that of `kong.log()`. For
            example, the following call:
            ``` lua
             kong.log.err("hello ", "world")
            ```
            would, within the core, produce a log line similar to:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```
            If invoked from within a plugin (for example, `key-auth`) it would include the
            namespace prefix:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 [key-auth] hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```

        Phases:
            init_worker, certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            kong.log.warn("something require attention")

            kong.log.err("something failed: ", err)

            kong.log.alert("something requires immediate action")

        :parameter *args: All params will be concatenated and stringified before being sent to the log.
        :type *args: Any

        :return: Throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def serialize() -> None:
        """

        """
        pass

    @staticmethod
    def set_serialize_value(key: str, value: Any, options: table) -> table:
        """

            Sets a value to be used on the `serialize` custom table.
            Logging plugins use the output of `kong.log.serialize()` as a base for their logs.
            This function lets you customize the log output.
            It can be used to replace existing values in the output, or to delete
            existing values by passing `nil`.
            **Note:** The type-checking of the `value` parameter can take some time, so
            it is deferred to the `serialize()` call, which happens in the log
            phase in most real-usage cases.

        Phases:
            certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            # Adds a new value to the serialized table

            kong.log.set_serialize_value("my_new_value", 1)

            assert(kong.log.serialize().my_new_value == 1)

            # Value can be a table

            kong.log.set_serialize_value("my", { new = { value = 2 } })

            assert(kong.log.serialize().my.new.value == 2)

            # It is possible to change an existing serialized value

            kong.log.set_serialize_value("my_new_value", 3)

            assert(kong.log.serialize().my_new_value == 3)

            # Unset an existing value by setting it to nil

            kong.log.set_serialize_value("my_new_value", nil)

            assert(kong.log.serialize().my_new_value == nil)

            # Dots in the key are interpreted as table accesses

            kong.log.set_serialize_value("my.new.value", 4)

            assert(kong.log.serialize().my.new_value == 4)

        :parameter key: The name of the field.
        :type key: str
        :parameter value: Value to be set. When a table is used, its keys must be numbers, strings, or booleans, and its values can be numbers, strings, or other tables like itself, recursively.
        :type value: Any
        :parameter options: Can contain two entries: options.mode can be `set` (the default, always sets), `add` (only add if entry does not already exist) and `replace` (only change value if it already exists).
        :type options: table

        :return: The request information table.

        :rtype: table
        """
        pass

    @staticmethod
    def warn(*args: Any) -> None:
        """

            Similar to `kong.log()`, but the produced log has the severity given by
            `<level>`, instead of `notice`. The supported levels are:
            * `kong.log.alert()`
            * `kong.log.crit()`
            * `kong.log.err()`
            * `kong.log.warn()`
            * `kong.log.notice()`
            * `kong.log.info()`
            * `kong.log.debug()`
            Logs have the same format as that of `kong.log()`. For
            example, the following call:
            ``` lua
             kong.log.err("hello ", "world")
            ```
            would, within the core, produce a log line similar to:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```
            If invoked from within a plugin (for example, `key-auth`) it would include the
            namespace prefix:
            ``` plain
            2017/07/09 19:36:25 [error] 25932#0: *1 [kong] some_file.lua:54 [key-auth] hello world, client: 127.0.0.1, server: localhost, request: "GET /log HTTP/1.1", host: "localhost"
            ```

        Phases:
            init_worker, certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            kong.log.warn("something require attention")

            kong.log.err("something failed: ", err)

            kong.log.alert("something requires immediate action")

        :parameter *args: All params will be concatenated and stringified before being sent to the log.
        :type *args: Any

        :return: Throws an error on invalid inputs.

        :rtype: None
        """
        pass

    pass


import os
import re
import sys
import time
import traceback
import threading
import msgpack

from .const import PY3K
if PY3K:
    from socketserver import ThreadingMixIn, UnixStreamServer as sUnixStreamServer
else:
    from SocketServer import ThreadingMixIn, UnixStreamServer as sUnixStreamServer

from gevent import socket as gsocket, sleep as gsleep, spawn as gspawn
from gevent.server import StreamServer as gStreamServer

from .exception import PluginServerException

cmdre = re.compile("([a-z])([A-Z])")

DEFAULT_SOCKET_NAME = "python_pluginserver.sock"

def write_response(fd, msgid, response):
    fd.send(msgpack.packb([
        1,  # is response
        msgid,
        None,
        response
    ]))

def write_error(fd, msgid, error):
    fd.send(msgpack.packb([
        1,  # is response
        msgid,
        error,
        None
    ]))


class WrapSocket(object):
    def __init__(self, socket):
        self.socket = socket

    def read(self, n):
        return self.socket.recv(n)

class Server(object):
    def __init__(self, plugin_server):
        self.ps = plugin_server
        self.logger = plugin_server.logger

    def handle(self, fd, address, *_):
        # can't use socket.makefile here, since it returns a blocking IO
        # msgpack.Unpacker only expects read() but no other semantics to exist
        sockf = WrapSocket(fd)
        unpacker = msgpack.Unpacker(sockf)

        for _, msgid, method, args in unpacker:
            ns, cmd = method.split(".")
            if ns != "plugin":
                write_error(fd, msgid, "RPC for %s is not supported" % ns)
                continue

            cmd_r = cmd[0].lower() + cmdre.sub(lambda m: "%s_%s" % (m.group(1), m.group(2).lower()), cmd[1:])
            try:
                self.logger.debug("rpc: #%d method: %s args: %s" % (msgid, method, args))
                ret = getattr(self.ps, cmd_r)(*args)
                self.logger.debug("rpc: #%d return: %s" % (msgid, ret))
                write_response(fd, msgid, ret)
            except PluginServerException as ex:
                self.logger.warn("rpc: #%d error: %s" % (msgid, str(ex)))
                write_error(fd, msgid, str(ex))
            except Exception as ex:
                self.logger.error("rpc: #%d exception: %s" % (msgid, traceback.format_exc()))
                write_error(fd, msgid, str(ex))

class tUnixStreamServer(ThreadingMixIn, sUnixStreamServer):
    pass

def watchdog(sleep, logger):
    while True:
        if os.getppid() == 1:  # parent dead, process adopted by init
            logger.info("Kong exits, terminating...")
            sys.exit()
        sleep(1)

class UnixStreamServer(Server):
    def __init__(self, pluginserver, path,
                 sock_name=DEFAULT_SOCKET_NAME, use_gevent=True, listen_queue_size=4096):
        Server.__init__(self, pluginserver)
        self.path = os.path.join(path, sock_name)
        self.use_gevent = use_gevent
        self.listen_queue_size = listen_queue_size

    def serve_forever(self):
        if os.path.exists(self.path):
            os.remove(self.path)

        if self.use_gevent:
            listener = gsocket.socket(gsocket.AF_UNIX, gsocket.SOCK_STREAM)
            listener.bind(self.path)
            listener.listen(self.listen_queue_size)

            self.logger.info("server (gevent) started at path " + self.path)

            gspawn(watchdog, gsleep, self.logger)

            gStreamServer(listener, self.handle).serve_forever()
        else:
            import socket
            self.logger.info("server started at path " + self.path)

            t = threading.Thread(
                target=watchdog,
                args=(time.sleep, self.logger, ),
            )
            t.setDaemon(True)
            t.start()
            s = tUnixStreamServer(self.path, self.handle, bind_and_activate=False)
            s.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
            s.server_bind()
            s.socket.listen(self.listen_queue_size)
            s.serve_forever()



# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/service/request.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class request():


    @staticmethod
    def add_header(header: str, value: Any) -> None:
        """

            Adds a request header with the given value to the request to the Service. Unlike
            `kong.service.request.set_header()`, this function doesn't remove any existing
            headers with the same name. Instead, several occurrences of the header will be
            present in the request. The order in which headers are added is retained.

        Phases:
            rewrite`, access`

        Example:
            kong.service.request.add_header("Cache-Control", "no-cache")

            kong.service.request.add_header("Cache-Control", "no-store")

        :parameter header: The header name. Example: "Cache-Control".
        :type header: str
        :parameter value: The header value. Example: "no-cache".
        :type value: Any

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def clear_header(header: str) -> None:
        """

            Removes all occurrences of the specified header from the request to the Service.

        Phases:
            rewrite`, access`

        Example:
            kong.service.request.set_header("X-Foo", "foo")

            kong.service.request.add_header("X-Foo", "bar")

            kong.service.request.clear_header("X-Foo")

            # from here onwards, no X-Foo headers will exist in the request

        :parameter header: The header name. Example: "X-Foo".
        :type header: str

        :return: throws an error on invalid inputs.
            The function does not throw an error if no header was removed.

        :rtype: None
        """
        pass

    @staticmethod
    def disable_tls() -> Tuple[bool, str]:
        """

            Disables the TLS handshake to upstream for [ngx\_stream\_proxy\_module](https://nginx.org/en/docs/stream/ngx_stream_proxy_module.html).
            This overrides the [proxy\_ssl](https://nginx.org/en/docs/stream/ngx_stream_proxy_module.html#proxy_ssl) directive, effectively setting it to `off`
            for the current stream session.
            Once this function has been called, it is not possible to re-enable TLS handshake for the current session.

        Phases:
            preread`, balancer`

        Example:
            ok, err = kong.service.request.disable_tls()

            if not ok:

                # do something with error

        :return: `true` if the operation succeeded, `nil` if an error occurred.

        :rtype: bool
        :return: An error message describing the error if there was one.

        :rtype: str
        """
        pass

    @staticmethod
    def enable_buffering() -> None:
        """

            Enables buffered proxying, which allows plugins to access Service body and
            response headers at the same time.

        Phases:
            rewrite`, access`

        Example:
            kong.service.request.enable_buffering()

        :return: 

        :rtype: None
        """
        pass

    @staticmethod
    def set_body(args: table, mimetype: Optional[str]) -> Tuple[bool, str]:
        """

            Sets the body of the request to the Service. Unlike
            `kong.service.request.set_raw_body()`, the `args` argument must be a table, and
            is encoded with a MIME type.  The encoding MIME type can be specified in
            the optional `mimetype` argument, or if left unspecified, is chosen based
            on the `Content-Type` header of the client's request.
            Behavior based on MIME type in the `Content-Type` header:
            * `application/x-www-form-urlencoded`: Encodes the arguments as
              form-encoded. Keys are produced in lexicographical
              order. The order of entries within the same key (when values are
              given as an array) is retained. Any string values given are URL-encoded.
            * `multipart/form-data`: Encodes the arguments as multipart form data.
            * `application/json`: Encodes the arguments as JSON (same as
              `kong.service.request.set_raw_body(json.encode(args))`). Lua types are
              converted to matching JSON types.
            If the MIME type is none of the above, this function returns `nil` and
            an error message indicating the body could not be encoded.
            If the `mimetype` argument is specified, the `Content-Type` header is
            set accordingly in the request to the Service.
            If further control of the body generation is needed, a raw body can be given as
            a string with `kong.service.request.set_raw_body()`.

        Phases:
            rewrite`, access`

        Example:
            kong.service.set_header("application/json")

            ok, err = kong.service.request.set_body({

            name = "John Doe",

            age = 42,

            numbers = {1, 2, 3}

            })

            # Produces the following JSON body:

            # { "name": "John Doe", "age": 42, "numbers":[1, 2, 3] }

            ok, err = kong.service.request.set_body({

            foo = "hello world",

            bar = {"baz", "bla", true},

            zzz = true,

            blo = ""

            }, "application/x-www-form-urlencoded")

            # Produces the following body:

            # bar=baz&bar=bla&bar&blo=&foo=hello%20world&zzz

        :parameter args: A table with data to be converted to the appropriate format
            and stored in the body.
        :type args: table
        :parameter mimetype: can be one of:
        :type mimetype: str

        :return: `true` on success, `nil` otherwise.

        :rtype: bool
        :return: `nil` on success, an error message in case of error.
            Throws an error on invalid inputs.

        :rtype: str
        """
        pass

    @staticmethod
    def set_header(header: str, value: Any) -> None:
        """

            Sets a header in the request to the Service with the given value. Any existing header
            with the same name will be overridden.
            If the `header` argument is `"host"` (case-insensitive), then this also
            sets the SNI of the request to the Service.

        Phases:
            rewrite`, access`, balancer`

        Example:
            kong.service.request.set_header("X-Foo", "value")

        :parameter header: The header name. Example: "X-Foo".
        :type header: str
        :parameter value: The header value. Example: "hello world".
        :type value: Any

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def set_headers(headers: table) -> None:
        """

            Sets the headers of the request to the Service. Unlike
            `kong.service.request.set_header()`, the `headers` argument must be a table in
            which each key is a string (corresponding to a header's name), and each value
            is a string, or an array of strings.
            The resulting headers are produced in lexicographical order. The order of
            entries with the same name (when values are given as an array) is retained.
            This function overrides any existing header bearing the same name as those
            specified in the `headers` argument. Other headers remain unchanged.
            If the `"Host"` header is set (case-insensitive), then this also sets
            the SNI of the request to the Service.

        Phases:
            rewrite`, access`

        Example:
            kong.service.request.set_header("X-Foo", "foo1")

            kong.service.request.add_header("X-Foo", "foo2")

            kong.service.request.set_header("X-Bar", "bar1")

            kong.service.request.set_headers({

            ["X-Foo"] = "foo3",

            ["Cache-Control"] = { "no-store", "no-cache" },

            ["Bla"] = "boo"

            })

            # Will add the following headers to the request, in this order:

            # X-Bar: bar1

            # Bla: boo

            # Cache-Control: no-store

            # Cache-Control: no-cache

            # X-Foo: foo3

        :parameter headers: A table where each key is a string containing a header name
            and each value is either a string or an array of strings.
        :type headers: table

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def set_method(method: str) -> None:
        """

            Sets the HTTP method for the request to the service.

        Phases:
            rewrite`, access`

        Example:
            kong.service.request.set_method("DELETE")

        :parameter method: The method string, which must be in all
            uppercase. Supported values are: `"GET"`, `"HEAD"`, `"PUT"`, `"POST"`,
            `"DELETE"`, `"OPTIONS"`, `"MKCOL"`, `"COPY"`, `"MOVE"`, `"PROPFIND"`,
            `"PROPPATCH"`, `"LOCK"`, `"UNLOCK"`, `"PATCH"`, or `"TRACE"`.
        :type method: str

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def set_path(path: str) -> None:
        """

            Sets the path component for the request to the service.
            The input accepts any valid *normalized* URI (including UTF-8 characters)
            and this API will perform necessary escaping according to the RFC
            to make the request valid.
            Input should **not** include the query string.

        Phases:
            access`

        Example:
            kong.service.request.set_path("/v2/movies")

        :parameter path: The path string. Special characters and UTF-8
            characters are allowed, for example: `"/v2/movies"` or `"/foo/😀"`.
        :type path: str

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def set_query(args: table) -> None:
        """

            Set the query string of the request to the Service.
            Unlike `kong.service.request.set_raw_query()`, the `query` argument must be a
            table in which each key is a string (corresponding to an argument's name), and
            each value is either a boolean, a string, or an array of strings or booleans.
            Additionally, all string values will be URL-encoded.
            The resulting query string contains keys in their lexicographical order. The
            order of entries within the same key (when values are given as an array) is
            retained.
            If further control of the query string generation is needed, a raw query
            string can be given as a string with `kong.service.request.set_raw_query()`.

        Phases:
            rewrite`, access`

        Example:
            kong.service.request.set_query({

            foo = "hello world",

            bar = {"baz", "bla", true},

            zzz = true,

            blo = ""

            })

            # Produces the following query string:

            # bar=baz&bar=bla&bar&blo=&foo=hello%20world&zzz

        :parameter args: A table where each key is a string (corresponding to an
            argument name), and each value is either a boolean, a string, or an array of
            strings or booleans. Any string values given are URL-encoded.
        :type args: table

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def set_raw_body(body: str) -> None:
        """

            Sets the body of the request to the Service.
            The `body` argument must be a string and will not be processed in any way.
            This function also sets the `Content-Length` header appropriately. To set an
            empty body, you can provide an empty string (`""`) to this function.
            For a higher-level function to set the body based on the request content type,
            see `kong.service.request.set_body()`.

        Phases:
            rewrite`, access`

        Example:
            kong.service.request.set_raw_body("Hello, world!")

        :parameter body: The raw body.
        :type body: str

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def set_raw_query(query: str) -> None:
        """

            Sets the query string of the request to the Service. The `query` argument is a
            string (without the leading `?` character), and is not processed in any
            way.
            For a higher-level function to set the query string from a Lua table of
            arguments, see `kong.service.request.set_query()`.

        Phases:
            rewrite`, access`

        Example:
            kong.service.request.set_raw_query("zzz&bar=baz&bar=bla&bar&blo=&foo=hello%20world")

        :parameter query: The raw querystring. Example:
            `"foo=bar&bla&baz=hello%20world"`.
        :type query: str

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def set_scheme(scheme: str) -> None:
        """

            Sets the protocol to use when proxying the request to the Service.

        Phases:
            access`

        Example:
            kong.service.request.set_scheme("https")

        :parameter scheme: The scheme to be used. Supported values are `"http"` or `"https"`.
        :type scheme: str

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    pass


# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/service/response.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class response():


    @staticmethod
    def get_body(mimetype: Optional[str], max_args: Optional[number]) -> str:
        """

            Returns the decoded buffered body.

        Phases:
            header_filter`, body_filter`, log`

        Example:
            # Plugin needs to call kong.service.request.enable_buffering() on `rewrite`

            # or `access` phase prior calling this function.

            body = kong.service.response.get_body()

        :parameter mimetype: The MIME type of the response (if known).
        :type mimetype: str
        :parameter max_args: Sets a limit on the maximum number of (what?)
            that can be parsed.
        :type max_args: number

        :return: The raw buffered body

        :rtype: str
        """
        pass

    @staticmethod
    def get_header(name: str) -> str:
        """

            Returns the value of the specified response header.
            Unlike `kong.response.get_header()`, this function only returns a header
            if it is present in the response from the Service (ignoring headers added by Kong
            itself).

        Phases:
            header_filter`, body_filter`, log`

        Example:
            # Given a response with the following headers:

            # X-Custom-Header: bla

            # X-Another: foo bar

            # X-Another: baz

            kong.log.inspect(kong.service.response.get_header("x-custom-header")) # "bla"

            kong.log.inspect(kong.service.response.get_header("X-Another"))       # "foo bar"

        :parameter name: The name of the header.
            Header names in are case-insensitive and are normalized to lowercase, and
            dashes (`-`) can be written as underscores (`_`); that is, the header
            `X-Custom-Header` can also be retrieved as `x_custom_header`.
        :type name: str

        :return: The value of the header, or `nil` if a header with
            `name` is not found in the response. If a header with the same name is present
            multiple times in the response, this function returns the value of the
            first occurrence of this header.

        :rtype: str
        """
        pass

    @staticmethod
    def get_headers(max_headers: Optional[number]) -> Tuple[table, str]:
        """

            Returns a Lua table holding the headers from the Service response. Keys are
            header names. Values are either a string with the header value, or an array of
            strings if a header was sent multiple times. Header names in this table are
            case-insensitive and dashes (`-`) can be written as underscores (`_`); that is,
            the header `X-Custom-Header` can also be retrieved as `x_custom_header`.
            Unlike `kong.response.get_headers()`, this function only returns headers that
            are present in the response from the Service (ignoring headers added by Kong itself).
            If the request is not proxied to a Service (e.g. an authentication plugin rejected
            a request and produced an HTTP 401 response), then the returned `headers` value
            might be `nil`, since no response from the Service has been received.
            By default, this function returns up to **100** headers. The optional
            `max_headers` argument can be specified to customize this limit, but must be
            greater than **1** and not greater than **1000**.

        Phases:
            header_filter`, body_filter`, log`

        Example:
            # Given a response with the following headers:

            # X-Custom-Header: bla

            # X-Another: foo bar

            # X-Another: baz

            headers = kong.service.response.get_headers()

            if headers:

                kong.log.inspect(headers.x_custom_header) # "bla"

            kong.log.inspect(headers.x_another[1])    # "foo bar"

            kong.log.inspect(headers["X-Another"][2]) # "baz"

        :parameter max_headers: Sets a limit on the maximum number of
            headers that can be parsed.
        :type max_headers: number

        :return: The response headers in table form.

        :rtype: table
        :return: If more headers than `max_headers` are present, returns
            a string with the error `"truncated"`.

        :rtype: str
        """
        pass

    @staticmethod
    def get_raw_body() -> str:
        """

            Returns the raw buffered body.

        Phases:
            header_filter`, body_filter`, log`

        Example:
            # Plugin needs to call kong.service.request.enable_buffering() on `rewrite`

            # or `access` phase prior calling this function.

            body = kong.service.response.get_raw_body()

        :return: The raw buffered body.

        :rtype: str
        """
        pass

    @staticmethod
    def get_status() -> number:
        """

            Returns the HTTP status code of the response from the Service as a Lua number.

        Phases:
            header_filter`, body_filter`, log`

        Example:
            kong.log.inspect(kong.service.response.get_status()) # 418

        :return: The status code from the response from the Service, or `nil`
            if the request was not proxied (that is, if `kong.response.get_source()` returned
            anything other than `"service"`).

        :rtype: number
        """
        pass

    pass


import os
import sys
import time
from .const import PY3K, FILEPATH
from .exception import PDKException
if PY3K:
    import importlib.util

    def load_module(name, path):
        spec = importlib.util.spec_from_file_location(name, path)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        sys.modules[name] = mod
        return mod
else:
    import imp
    methods = {
        '.py': imp.load_source,
        '.pyc': imp.load_compiled,
        '.pyo': imp.load_compiled,
        '.pyd': imp.load_dynamic,
        '.so': imp.load_dynamic,
    }

    def load_module(name, path):
        ext = os.path.splitext(path)[1]
        mod = methods[ext](name, path)
        sys.modules[name] = mod
        return mod

phases = ("certificate", "rewrite", "log", "access", "preread", "response")

class Module(object):
    def __init__(self, name, path=None, module=None):
        self.name = name
        self.load_time = time.time()

        if path:
            mod = load_module(name, path)
            self.mod = mod
            self.mtime = os.stat(path).st_mtime
        elif module:
            mod = module
            self.mtime = os.stat(FILEPATH).st_mtime
        else:
            raise PDKException("either path or module needs to be passed in")

        self.cls = getattr(mod, "Plugin")

        self.phases = []
        for phase in phases:
            if hasattr(self.cls, phase):
                self.phases.append(phase)

        self.priority = 0
        self.version = None

        for attr in ('version', 'priority'):
            au = attr.upper()
            if hasattr(mod, au):
                setattr(self, attr, getattr(mod, au))

        if hasattr(mod, "Schema"):
            self.schema = mod.Schema
        else:
            self.schema = []

        self.last_start_instance_time = 0
        self.last_close_instance_time = 0

    def new(self, config):
        self.last_start_instance_time = time.time()
        return Instance(self.name, config, self.cls, self.set_last_close_instance_time)

    def set_last_close_instance_time(self):
        self.last_close_instance_time = time.time()


class Instance(object):
    def __init__(self, name, config, cls, close_cb):
        self.cls = cls(config)
        self.name = name
        self.config = config
        self.start_time = time.time()
        self.last_used_time = 0
        self.close_cb = close_cb

    def is_expired(self, ttl=60):
        until = time.time() - ttl
        return self.start_time < until and self.last_used_time < until

    def reset_expire_ts(self):
        self.last_used_time = time.time()
        return self.cls



# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/nginx.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str

from .shared import shared as cls_shared

class nginx():

    shared = cls_shared

    @staticmethod
    def get_ctx(k: str) -> Any:
        """

            get a key-value pair from Kong's per-request context

        :parameter k: key for the ctx data
        :type k: str

        :return: the per-request context data in ngx.ctx

        :rtype: Any
        """
        pass

    @staticmethod
    def get_subsystem() -> str:
        """

        :return: the subsystem name

        :rtype: str
        """
        pass

    @staticmethod
    def get_tls1_version_str() -> str:
        """

        :return: the TLSv1 version string

        :rtype: str
        """
        pass

    @staticmethod
    def get_var() -> str:
        """

        :return: the NGINX version string

        :rtype: str
        """
        pass

    @staticmethod
    def req_start_time() -> number:
        """

            get the current request's start timestamp

        :return: req_start_time

        :rtype: number
        """
        pass

    @staticmethod
    def set_ctx(k: str, any: str) -> None:
        """

            set a key-value pair in Kong's per-request context

        :parameter k: key for the ctx data
        :type k: str
        :parameter any: value for the ctx data
        :type any: str

        """
        pass

    pass


# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/client.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str

from .tls import tls as cls_tls

class client():

    tls = cls_tls

    @staticmethod
    def authenticate(consumer: table, credential: table) -> None:
        """

            Sets the authenticated consumer and/or credential for the current request.
            While both `consumer` and `credential` can be `nil`,
            at least one of them must exist. Otherwise, this function will throw an
            error.

        Phases:
            access

        Example:
            # assuming `credential` and `consumer` have been set by some authentication code

            kong.client.authenticate(consumer, credentials)

        :parameter consumer: The consumer to set. If no
            value is provided, then any existing value will be cleared.
        :type consumer: table
        :parameter credential: The credential to set. If
            no value is provided, then any existing value will be cleared.
        :type credential: table

        """
        pass

    @staticmethod
    def get_consumer() -> table:
        """

            Returns the `consumer` entity of the currently authenticated consumer.
            If not set yet, it returns `nil`.

        Phases:
            access, header_filter, response, body_filter, log

        Example:
            consumer = kong.client.get_consumer()

            if consumer:

                consumer_id = consumer.id

            else:

                # request not authenticated yet, or a credential

            # without a consumer (external auth)

        :return: The authenticated consumer entity.

        :rtype: table
        """
        pass

    @staticmethod
    def get_credential() -> str:
        """

            Returns the credentials of the currently authenticated consumer.
            If not set yet, it returns `nil`.

        Phases:
            access, header_filter, response, body_filter, log

        Example:
            credential = kong.client.get_credential()

            if credential:

                consumer_id = credential.consumer_id

            else:

                # request not authenticated yet

        :return: The authenticated credential.

        :rtype: str
        """
        pass

    @staticmethod
    def get_forwarded_ip() -> str:
        """

            Returns the remote address of the client making the request. Unlike
            `kong.client.get_ip`, this function will consider forwarded addresses in
            cases when a load balancer is in front of Kong. Whether this function
            returns a forwarded address or not depends on several Kong configuration
            parameters:
            * [trusted\_ips](https://docs.konghq.com/gateway/latest/reference/configuration/#trusted_ips)
            * [real\_ip\_header](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_header)
            * [real\_ip\_recursive](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_recursive)

        Phases:
            certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            # Given a client with IP 127.0.0.1 making connection through

            # a load balancer with IP 10.0.0.1 to Kong answering the request for

            # https://username:password@example.com:1234/v1/movies

            kong.client.get_forwarded_ip() # "127.0.0.1"

            # Note: This example assumes that 10.0.0.1 is one of the trusted IPs, and that

            # the load balancer adds the right headers matching with the configuration

            # of `real_ip_header`, e.g. `proxy_protocol`.

        :return: The remote IP address of the client making the request,
            considering forwarded addresses.

        :rtype: str
        """
        pass

    @staticmethod
    def get_forwarded_port() -> number:
        """

            Returns the remote port of the client making the request. Unlike
            `kong.client.get_port`, this function will consider forwarded ports in cases
            when a load balancer is in front of Kong. Whether this function returns a
            forwarded port or not depends on several Kong configuration parameters:
            * [trusted\_ips](https://docs.konghq.com/gateway/latest/reference/configuration/#trusted_ips)
            * [real\_ip\_header](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_header)
            * [real\_ip\_recursive](https://docs.konghq.com/gateway/latest/reference/configuration/#real_ip_recursive)

        Phases:
            certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            # [client]:40000 <-> 80:[balancer]:30000 <-> 80:[kong]:20000 <-> 80:[service]

            kong.client.get_forwarded_port() # 40000

            # Note: This example assumes that [balancer] is one of the trusted IPs, and that

            # the load balancer adds the right headers matching with the configuration

            # of `real_ip_header`, e.g. `proxy_protocol`.

        :return: The remote client port, considering forwarded ports.

        :rtype: number
        """
        pass

    @staticmethod
    def get_ip() -> str:
        """

            Returns the remote address of the client making the request. This module
            **always** returns the address of the client directly connecting to Kong.
            That is, in cases when a load balancer is in front of Kong, this function
            returns the load balancer's address, and **not** that of the
            downstream client.

        Phases:
            certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            # Given a client with IP 127.0.0.1 making connection through

            # a load balancer with IP 10.0.0.1 to Kong answering the request for

            # https://example.com:1234/v1/movies

            kong.client.get_ip() # "10.0.0.1"

        :return: The remote IP address of the client making the request.

        :rtype: str
        """
        pass

    @staticmethod
    def get_port() -> number:
        """

            Returns the remote port of the client making the request. This
            **always** returns the port of the client directly connecting to Kong. That
            is, in cases when a load balancer is in front of Kong, this function
            returns the load balancer's port, and **not** that of the downstream client.

        Phases:
            certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            # [client]:40000 <-> 80:[balancer]:30000 <-> 80:[kong]:20000 <-> 80:[service]

            kong.client.get_port() # 30000

        :return: The remote client port.

        :rtype: number
        """
        pass

    @staticmethod
    def get_protocol(allow_terminated: Optional[bool]) -> Tuple[str, err]:
        """

            Returns the protocol matched by the current route (`"http"`, `"https"`, `"tcp"` or
            `"tls"`), or `nil`, if no route has been matched, which can happen when dealing with
            erroneous requests.

        Phases:
            access, header_filter, response, body_filter, log

        Example:
            kong.client.get_protocol() # "http"

        :parameter allow_terminated: If set, the `X-Forwarded-Proto` header is checked when checking for HTTPS.
        :type allow_terminated: bool

        :return: Can be one of `"http"`, `"https"`, `"tcp"`, `"tls"` or `nil`.

        :rtype: str
        :return: `nil` if successful, or an error message if it fails.

        :rtype: err
        """
        pass

    @staticmethod
    def load_consumer(consumer_id: str, search_by_username: Optional[bool]) -> Tuple[table, err]:
        """

            Returns the consumer from the datastore.
            Looks up the consumer by ID, and can optionally do a second search by name.

        Phases:
            access, header_filter, response, body_filter, log

        Example:
            consumer_id = "john_doe"

            consumer = kong.client.load_consumer(consumer_id, true)

        :parameter consumer_id: The consumer ID to look up.
        :type consumer_id: str
        :parameter search_by_username: If truthy,
            and if the consumer is not found by ID,
            then a second search by username will be performed.
        :type search_by_username: bool

        :return: Consumer entity or `nil`.

        :rtype: table
        :return: `nil` if successful, or an error message if it fails.

        :rtype: err
        """
        pass

    pass


# coding:utf-8
# Contributor:
#      fffonion        <fffonion@gmail.com>

import os
import sys
import datetime
import locale
import traceback
from threading import RLock
from .const import PY3K

class tz_GMT8(datetime.tzinfo):
    def utcoffset(self, dt):
        return datetime.timedelta(hours=8)

    def dst(self, dt):
        return datetime.timedelta(0)


if PY3K:
    unicode = None

def safestr(s):
    if (PY3K and isinstance(s, bytes)) or (not PY3K and not isinstance(s, unicode)):
        s = s.decode("utf-8")
    if PY3K:
        # python<=3.5 hack
        if sys.version_info.minor <= 5:
            return s \
                .encode(locale.getdefaultlocale()[1] or 'utf-8', 'replace') \
                .decode(locale.getdefaultlocale()[1] or 'utf-8', 'replace')
        return s
    return s.encode(locale.getdefaultlocale()[1] or 'utf-8', 'replace')
    # return _.decode('utf-8') if PY3K else _


if os.name == 'nt':
    endl = '\r\n'
else:  # assume posix
    endl = '\n'

class Logger(object):
    # paste from goagent
    CRITICAL = 5
    FATAL = CRITICAL
    ERROR = 4
    WARNING = 3
    WARN = WARNING
    INFO = 2
    DEBUG = 1
    VERBOSE = 0

    def __init__(self, *args, **kwargs):
        # self.level = self.__class__.INFO
        self.logf = None
        self.__write = __write = lambda x: sys.stdout.write(safestr(x))
        self.isatty = getattr(sys.stdout, 'isatty', lambda: False)()
        self.__set_error_color = lambda: None
        self.__set_warning_color = lambda: None
        self.__set_debug_color = lambda: None
        self.__set_verbose_color = lambda: None
        self.__reset_color = lambda: None
        if self.isatty:
            if os.name == 'nt':
                self._nt_color_lock = RLock()
                import ctypes
                SetConsoleTextAttribute = ctypes.windll.kernel32.SetConsoleTextAttribute
                GetStdHandle = ctypes.windll.kernel32.GetStdHandle
                self.__set_error_color = lambda: (self._nt_color_lock.acquire(), SetConsoleTextAttribute(GetStdHandle(-11), 0x0C))
                self.__set_warning_color = lambda: (self._nt_color_lock.acquire(), SetConsoleTextAttribute(GetStdHandle(-11), 0x06))
                self.__set_debug_color = lambda: (self._nt_color_lock.acquire(), SetConsoleTextAttribute(GetStdHandle(-11), 0x02))
                self.__set_verbose_color = lambda: (self._nt_color_lock.acquire(), SetConsoleTextAttribute(GetStdHandle(-11), 0x08))
                self.__set_bright_color = lambda: (self._nt_color_lock.acquire(), SetConsoleTextAttribute(GetStdHandle(-11), 0x0F))
                self.__reset_color = lambda: (SetConsoleTextAttribute(GetStdHandle(-11), 0x07), self._nt_color_lock.release())
            elif os.name == 'posix':
                self.__set_error_color = lambda: __write('\033[31m')
                self.__set_warning_color = lambda: __write('\033[33m')
                self.__set_debug_color = lambda: __write('\033[32m')
                self.__set_verbose_color = lambda: __write('\033[36m')
                self.__set_bright_color = lambda: __write('\033[32m')
                self.__reset_color = lambda: __write('\033[0m')

    @classmethod
    def getLogger(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def cleanup(self):
        if self.logf:
            _ = self.logf
            self.logf = None
            _.close()

    def set_logfile(self, fpath):
        if self.logf:
            self.logf.close()
        self.logf = open(fpath, "ab")

    def set_level(self, level):
        f = ('verbose', 'debug', 'info')
        lv = min(max(level, 0), 3)
        for p in range(lv):
            setattr(self, f[p], self.dummy)

    def log(self, level, fmt, *args, **kwargs):
        # fmt=du8(fmt)
        try:
            try:
                self.__write('%-4s - [%s] %s\n' % (level, datetime.datetime.now(tz_GMT8()).strftime('%X'), fmt % args))
            except (ValueError, TypeError):
                fmt = fmt.replace('%', '%%')
                self.__write('%-4s - [%s] %s\n' % (level, datetime.datetime.now(tz_GMT8()).strftime('%X'), fmt % args))
        except IOError:  # fix for Windows console
            pass
        sys.stdout.flush()
        if self.logf:
            _ = ('[%s] %s%s' % (datetime.datetime.now(tz_GMT8()).strftime('%b %d %X'), fmt % args, endl))
            self.logf.write(_.encode("utf-8", 'replace'))

    def dummy(self, *args, **kwargs):
        pass

    def debug(self, fmt, *args, **kwargs):
        self.__set_debug_color()
        self.log('DEBG', fmt, *args, **kwargs)
        self.__reset_color()

    def info(self, fmt, *args, **kwargs):
        self.log('INFO', fmt, *args)

    def verbose(self, fmt, *args, **kwargs):
        self.__set_verbose_color()
        self.log('VERB', fmt, *args, **kwargs)
        self.__reset_color()

    def warning(self, fmt, *args, **kwargs):
        self.__set_warning_color()
        self.log('WARN', fmt, *args, **kwargs)
        self.__reset_color()

    def warn(self, fmt, *args, **kwargs):
        self.warning(fmt, *args, **kwargs)

    def error(self, fmt, *args, **kwargs):
        self.__set_error_color()
        self.log('ERROR', fmt, *args, **kwargs)
        self.__reset_color()

    def exception(self, fmt, *args, **kwargs):
        self.error(fmt, *args, **kwargs)
        traceback.print_exc(file=sys.stderr)

    def critical(self, fmt, *args, **kwargs):
        self.__set_error_color()
        self.log('CRITICAL', fmt, *args, **kwargs)
        self.__reset_color()



# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/service.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str

from .request import request as cls_request
from .response import response as cls_response

class service():

    request = cls_request
    response = cls_response

    @staticmethod
    def set_target(host: str, port: number) -> None:
        """

            Sets the host and port on which to connect to for proxying the request.
            Using this method is equivalent to ask Kong to not run the load-balancing
            phase for this request, and consider it manually overridden.
            Load-balancing components such as retries and health-checks will also be
            ignored for this request.
            The `host` argument expects a string containing the IP address of the
            upstream server (IPv4/IPv6), and the `port` argument must contain a number
            representing the port on which to connect to.

        Phases:
            access

        Example:
            kong.service.set_target("service.local", 443)

            kong.service.set_target("192.168.130.1", 80)

        :parameter host: 
        :type host: str
        :parameter port: 
        :type port: number

        """
        pass

    @staticmethod
    def set_tls_verify(on: bool) -> Tuple[bool, str]:
        """

            Sets whether TLS verification is enabled while handshaking with the Service.
            The `on` argument is a boolean flag, where `true` means upstream verification
            is enabled and `false` disables it.
            This call affects only the current request. If the trusted certificate store is
            not set already (via [proxy_ssl_trusted_certificate](https://nginx.org/en/docs/http/ngx_http_proxy_module.html#proxy_ssl_trusted_certificate)
            or [kong.service.set_upstream_ssl_trusted_store](#kongserviceset_upstream_ssl_trusted_store)),
            then TLS verification will always fail with "unable to get local issuer certificate" error.

        Phases:
            rewrite`, access`, balancer`

        Example:
            ok, err = kong.service.set_tls_verify(true)

            if not ok:

                # do something with error

        :parameter on: Whether to enable TLS certificate verification for the current request
        :type on: bool

        :return: `true` if the operation succeeded, `nil` if an error occurred

        :rtype: bool
        :return: An error message describing the error if there was one

        :rtype: str
        """
        pass

    @staticmethod
    def set_tls_verify_depth(depth: number) -> Tuple[bool, str]:
        """

            Sets the maximum depth of verification when validating upstream server's TLS certificate.
            This call affects only the current request. For the depth to be actually used the verification
            has to be enabled with either the [proxy_ssl_verify](https://nginx.org/en/docs/http/ngx_http_proxy_module.html#proxy_ssl_verify)
            directive or using the [kong.service.set_tls_verify](#kongserviceset_tls_verify) function.

        Phases:
            rewrite`, access`, balancer`

        Example:
            ok, err = kong.service.set_tls_verify_depth(3)

            if not ok:

                # do something with error

        :parameter depth: Depth to use when validating. Must be non-negative
        :type depth: number

        :return: `true` if the operation succeeded, `nil` if an error occurred

        :rtype: bool
        :return: An error message describing the error if there was one

        :rtype: str
        """
        pass

    @staticmethod
    def set_upstream(host: str) -> Tuple[bool, str]:
        """

            Sets the desired Upstream entity to handle the load-balancing step for
            this request. Using this method is equivalent to creating a Service with a
            `host` property equal to that of an Upstream entity (in which case, the
            request would be proxied to one of the Targets associated with that
            Upstream).
            The `host` argument should receive a string equal to that of one of the
            Upstream entities currently configured.

        Phases:
            access

        Example:
            ok, err = kong.service.set_upstream("service.prod")

            if not ok:

                kong.log.err(err)

            return

        :parameter host: 
        :type host: str

        :return: `true` on success, or `nil` if no upstream entities
            where found

        :rtype: bool
        :return: An error message describing the error if there was
            one.

        :rtype: str
        """
        pass

    pass


# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/client/tls.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class tls():


    @staticmethod
    def disable_session_reuse() -> Tuple[bool, err]:
        """

            Prevents the TLS session for the current connection from being reused
            by disabling the session ticket and session ID for the current TLS connection.

        Phases:
            certificate

        Example:
            res, err = kong.client.tls.disable_session_reuse()

            if not res:

                # do something with err

        :return: Returns `true` if successful, `nil` if it fails.

        :rtype: bool
        :return: Returns `nil` if successful, or an error message if it fails.

        :rtype: err
        """
        pass

    @staticmethod
    def get_full_client_certificate_chain() -> Tuple[str, err]:
        """

            Returns the PEM encoded downstream client certificate chain with the
            client certificate at the top and intermediate certificates
            (if any) at the bottom.

        Phases:
            rewrite, access, balancer, header_filter, body_filter, log

        Example:
            cert, err = kong.client.get_full_client_certificate_chain()

            if err:

                # do something with errif not cert:

                # client did not complete mTLS# do something with cert

        :return: Returns a PEM-encoded client certificate if the mTLS
            handshake was completed, or `nil` if an error occurred or the client did
            not present its certificate.

        :rtype: str
        :return: Returns `nil` if successful, or an error message if it fails.

        :rtype: err
        """
        pass

    @staticmethod
    def request_client_certificate() -> Tuple[bool, err]:
        """

            Requests the client to present its client-side certificate to initiate mutual
            TLS authentication between server and client.
            This function *requests*, but does not *require* the client to start
            the mTLS process. The TLS handshake can still complete even if the client
            doesn't present a client certificate. However, in that case, it becomes a
            TLS connection instead of an mTLS connection, as there is no mutual
            authentication.
            To find out whether the client honored the request, use
            `get_full_client_certificate_chain` in later phases.

        Phases:
            certificate

        Example:
            res, err = kong.client.tls.request_client_certificate()

            if not res:

                # do something with err

        :return: Returns `true` if request is received, or `nil` if
            request fails.

        :rtype: bool
        :return: Returns `nil` if the handshake is successful, or an error
            message if it fails.

        :rtype: err
        """
        pass

    @staticmethod
    def set_client_verify() -> None:
        """

            Overrides the client's verification result generated by the log serializer.
            By default, the `request.tls.client_verify` field inside the log
            generated by Kong's log serializer is the same as the
            [$ssl_client_verify](https://nginx.org/en/docs/http/ngx_http_ssl_module.html#var_ssl_client_verify)
            Nginx variable.
            Only `"SUCCESS"`, `"NONE"`, or `"FAILED:<reason>"` are accepted values.
            This function does not return anything on success, and throws a Lua error
            in case of a failure.

        Phases:
            rewrite, access, balancer

        Example:
            kong.client.tls.set_client_verify("FAILED:unknown CA")

        """
        pass

    pass


# coding:utf-8
# constants module
# Contributor:
#      fffonion        <fffonion@gmail.com>

import os
import sys
import locale

PY3K = sys.version_info.major == 3
IRONPYTHON = sys.platform == 'cli'
EXEBUNDLE = getattr(sys, 'frozen', False)
LOCALE = locale.getdefaultlocale()[0]
CODEPAGE = locale.getdefaultlocale()[1] or 'ascii'

__version__ = 0.30

# https://github.com/soimort/you-get/you-get
if getattr(sys, 'frozen', False):
    # The application is frozen
    FILEPATH = os.path.dirname(os.path.realpath(sys.executable))
else:
    # The application is not frozen
    # Change this bit to match where you store your data files:
    FILEPATH = sys.path[0]



# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/router.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class router():


    @staticmethod
    def get_route() -> table:
        """

            Returns the current `route` entity. The request is matched against this
            route.

        Phases:
            access, header_filter, response, body_filter, log

        Example:
            route = kong.router.get_route()

            protocols = route.protocols

        :return: The `route` entity.

        :rtype: table
        """
        pass

    @staticmethod
    def get_service() -> table:
        """

            Returns the current `service` entity. The request is targeted to this
            upstream service.

        Phases:
            access, header_filter, response, body_filter, log

        Example:
            if kong.router.get_service():

                # routed by route & service entities

            else:

                # routed by a route without a service

        :return: The `service` entity.

        :rtype: table
        """
        pass

    pass


# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/node.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class node():


    @staticmethod
    def get_hostname() -> str:
        """

            Returns the name used by the local machine.

        Example:
            hostname = kong.node.get_hostname()

        :return: The local machine hostname.

        :rtype: str
        """
        pass

    @staticmethod
    def get_id() -> str:
        """

            Returns the ID used by this node to describe itself.

        Example:
            id = kong.node.get_id()

        :return: The v4 UUID used by this node as its ID.

        :rtype: str
        """
        pass

    @staticmethod
    def get_memory_stats(unit: Optional[str], scale: Optional[number]) -> table:
        """

            Returns memory usage statistics about this node.

        Example:
            res = kong.node.get_memory_stats()

            # res will have the following structure:

            {

            lua_shared_dicts = {

            kong = {

            allocated_slabs = 12288,

            capacity = 24576

            },

            kong_db_cache = {

            allocated_slabs = 12288,

            capacity = 12288

            }

            },

            workers_lua_vms = {

            {

            http_allocated_gc = 1102,

            pid = 18004

            },

            {

            http_allocated_gc = 1102,

            pid = 18005

            }

            }

            }

            res = kong.node.get_memory_stats("k", 1)

            # res will have the following structure:

            {

            lua_shared_dicts = {

            kong = {

            allocated_slabs = "12.0 KiB",

            capacity = "24.0 KiB",

            },

            kong_db_cache = {

            allocated_slabs = "12.0 KiB",

            capacity = "12.0 KiB",

            }

            },

            workers_lua_vms = {

            {

            http_allocated_gc = "1.1 KiB",

            pid = 18004

            },

            {

            http_allocated_gc = "1.1 KiB",

            pid = 18005

            }

            }

            }

        :parameter unit: The unit that memory is reported in. Can be
            any of `b/B`, `k/K`, `m/M`, or `g/G` for bytes, kibibytes, mebibytes,
            or gibibytes, respectively. Defaults to `b` (bytes).
        :type unit: str
        :parameter scale: The number of digits to the right of the decimal
            point. Defaults to 2.
        :type scale: number

        :return: A table containing memory usage statistics for this node.
            If `unit` is `b/B` (the default), reported values are Lua numbers.
            Otherwise, reported values are strings with the unit as a suffix.

        :rtype: table
        """
        pass

    pass


# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/ip.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class ip():


    @staticmethod
    def is_trusted(address: str) -> bool:
        """

            Depending on the `trusted_ips` configuration property,
            this function returns whether a given IP is trusted or not.
            Both ipv4 and ipv6 are supported.

        Phases:
            init_worker, certificate, rewrite, access, header_filter, response, body_filter, log

        Example:
            if kong.ip.is_trusted("1.1.1.1"):

                kong.log("The IP is trusted")

        :parameter address: A string representing an IP address.
        :type address: str

        :return: `true` if the IP is trusted, `false` otherwise.

        :rtype: bool
        """
        pass

    pass


from telnetlib import theNULL
from ..exception import PDKException

class FakeClasses(object):
    def __init__(self, prefix, call):
        self.prefix = prefix
        self.call = call

    def __call__(self, *a):
        return self.call(self.prefix, *a)

    # TODO
    def __str__(self):
        return self.call(self.prefix)

    def __getattr__(self, k):
        return FakeClasses(self.prefix + "." + k, self.call)

# those methods never return, instead, they exit from current request immediately
non_return_methods = set((
    "kong.response.exit",
    "kong.response.error",
))

def rpc_of(ch, lua_style):
    def f(m, *a):
        ch.put({
            "Method": m,
            "Args": a,
        })
        if m in non_return_methods:
            return

        data, err = ch.get()
        if lua_style:
            return data, err
        if err:
            raise PDKException("exception from %s: %s" %(m, err))
        return data
    return f

class Kong(object):
    def __init__(self, ch, lua_style=False):
        self.kong = FakeClasses("kong", self.bridge)
        self.rpc = rpc_of(ch, lua_style)

    def bridge(self, attr, *args):
        return self.rpc(attr, *args)



# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/nginx/shared.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class shared():


    @staticmethod
    def get(k: str) -> Any:
        """

            get a key-value pair from Kong's shared memory

        :parameter k: key for the ctx data
        :type k: str

        :return: the per-request context data in ngx.ctx

        :rtype: Any
        """
        pass

    @staticmethod
    def set(k: str, v: str) -> None:
        """

            set a key-value pair in Kong's shared memory

        :parameter k: key for the ctx data
        :type k: str
        :parameter v: value for the ctx data
        :type v: str

        """
        pass

    pass


import os
import time
import json

from .const import PY3K

import threading
if PY3K:
    from queue import Queue
else:
    from Queue import Queue

import multiprocessing
# patch connection API so it share with gevent.queue.Channel
if PY3K:
    from multiprocessing import connection
    connection.Connection.get = connection.Connection.recv
    connection.Connection.put = connection.Connection.send

from gevent import sleep as gsleep, spawn as gspawn
from gevent.lock import Semaphore as gSemaphore
from gevent.queue import Channel as gChannel

try:
    import setproctitle
except:  # noqa: E722 do not use bare 'except
    setproctitle = None

from .pdk import Kong
from .module import Module
from .exception import PluginServerException
from .logger import Logger

exts = ('.py', '.pyd', '.so')
entities = ('service', 'consumer', 'route', 'plugin', 'credential', 'memory_stats')

MSG_RET = 'ret'

def locked_by(lock_name):
    def f(fn):
        def wrapper(*args, **kwargs):
            self = args[0]
            lock = getattr(self, lock_name)
            lock.acquire()
            try:
                r = fn(*args, **kwargs)
            except Exception as ex:
                lock.release()
                raise(ex)
            lock.release()
            return r
        return wrapper

    return f

def _handler_event_func(cls_phase, ch, lua_style):
    cls_phase(Kong(ch, lua_style).kong)
    ch.put(MSG_RET)

def _multiprocessing_init(pool_name):
    p = multiprocessing.current_process()
    p.name = "%s: %s (ppid: %d)" % (pool_name, p.name, os.getppid())
    if setproctitle:
        setproctitle.setproctitle(p.name)

class PluginServer(object):
    def __init__(self, loglevel=Logger.WARNING, expire_ttl=60, plugin_dir=None,
                 use_multiprocess=False, use_gevent=False, name=None, lua_style=True):
        if use_multiprocess:
            sem = multiprocessing.Semaphore
        elif use_gevent:
            sem = gSemaphore
        else:
            sem = threading.Semaphore

        self.plugin_dir = plugin_dir
        self.plugins = {}
        self.instances = {}
        self.instance_id = 0
        self.i_lock = sem()
        self.events = {}
        self.event_id = 0
        self.e_lock = sem()

        self.logger = Logger()
        self.logger.set_level(loglevel)

        if plugin_dir:
            self._load_plugins()

        title = "Kong Python Plugin Server"
        if name:
            title = "%s \"%s\"" % (title, name)

        self.use_multiprocess = use_multiprocess
        self.use_gevent = use_gevent

        self.lua_style = lua_style

        if use_multiprocess:
            if not PY3K:
                raise NotImplementedError("multiprocessing mode is only supported in Python3")

            self._process_pool = multiprocessing.Pool(
                os.cpu_count(),
                _multiprocessing_init,
                (title, ),
            )
            self.logger.debug("plugin server is in multiprocessing mode")

        # start cleanup timer
        if use_gevent:
            self.logger.debug("plugin server is in gevent mode")
            # gspawn(self._clear_expired_plugins, expire_ttl)
        else:
            pass
            # t = threading.Thread(
            #     target=self._clear_expired_plugins,
            #     args=(expire_ttl, ),
            # )
            # t.setDaemon(True)
            # t.start()

        if setproctitle:
            ppid = os.getppid()
            if use_multiprocess:
                setproctitle.setproctitle("%s: Manager (ppid: %d)" % (title, ppid))
            else:
                setproctitle.setproctitle("%s (ppid: %d)" % (title, ppid))

    def _clear_expired_plugins(self, ttl):
        while True:
            if self.use_gevent:
                gsleep(ttl)
            else:
                time.sleep(ttl)

            self.i_lock.acquire()
            keys = list(self.instances.keys())
            for iid in keys:
                instance = self.instances[iid]
                if instance.is_expired(ttl):
                    self.logger.debug("cleanup instance #%d of %s" % (iid, instance.name))
                    del(self.instances[iid])
            self.i_lock.release()

    def cleanup(self):
        if self.use_multiprocess:
            self._process_pool.terminate()
            self._process_pool.join()

    def _load_plugins(self):
        if not self.plugin_dir:
            raise PluginServerException("plugin server is not initialized, call SetPluginDir first")

        for p in os.listdir(self.plugin_dir):
            n, ext = os.path.splitext(p)
            if ext in exts:
                path = os.path.join(self.plugin_dir, p)
                try:
                    mod = Module(n, path=path)
                except Exception as ex:
                    self.logger.warn("error loading plugin \"%s\": %s" % (n, ex))
                else:
                    self.logger.debug("loaded plugin \"%s\" from %s" % (n, path))
                    self.plugins[n] = mod

    def set_plugin_dir(self, dir):
        if not os.path.exists(dir):
            raise PluginServerException("%s not exists" % dir)
        self.plugin_dir = dir
        return "ok"

    @locked_by("i_lock")
    def get_status(self, *_):
        plugin_status = {}
        for name in self.plugins:
            instances = []
            for iid in self.instances:
                if self.instances[iid].name == name:
                    i = self.instance_status(iid)
                    instances.append(i)
            plugin = self.plugins[name]
            plugin_status[name] = {
                "Name": name,
                "Modtime": plugin.mtime,
                "LoadTime": plugin.load_time,
                "Instances": instances,
                "LastStartInstance": plugin.last_start_instance_time,
                "LastCloseInstance": plugin.last_close_instance_time,
            }
        return {
            "Pid": os.getpid(),
            "Plugins": plugin_status,
        }

    def get_plugin_info(self, name):
        if name not in self.plugins:
            raise PluginServerException("%s not initizlied" % name)

        plugin = self.plugins[name]

        info = {
            "Name": name,
            "Phases": plugin.phases,
            "Priority": plugin.priority,
            "Schema": {
                "name": name,
                "fields": [{
                    "config": {
                        "type": "record",
                        "fields": plugin.schema,
                    }
                }],
            },
        }
        return info

    @locked_by("i_lock")
    def start_instance(self, cfg):
        name = cfg['Name']
        if name not in self.plugins:
            raise PluginServerException("%s not initizlied" % name)
        plugin = self.plugins[name]

        config = json.loads(cfg['Config'])
        iid = self.instance_id
        self.instances[iid] = plugin.new(config)
        self.instance_id = iid + 1

        self.logger.info("instance #%d of %s started" % (iid, name))

        return {
            "Name": name,
            "Id": iid,
            "Config": config,
            "StartTime": time.time()
        }

    def instance_status(self, iid):
        if iid not in self.instances:
            # Note: Kong expect the error to start with "no plugin instance"
            raise PluginServerException("no plugin instance #%s" % iid)

        ins = self.instances[iid]

        return {
            "Name": ins.name,
            "Id": iid,
            "Config": ins.config,
            "StartTime": ins.start_time,
        }

    @locked_by("i_lock")
    def close_instance(self, iid):
        if iid not in self.instances:
            # Note: Kong expect the error to start with "no plugin instance"
            raise PluginServerException("no plugin instance #%s" % iid)

        ins = self.instances[iid]
        ins.close_cb()
        del(self.instances[iid])

        return {
            "Name": ins['name'],
            "Id": iid,
            "Config": ins['config'],
        }

    @locked_by("e_lock")
    def handle_event(self, event):
        iid = event['InstanceId']
        if iid not in self.instances:
            # Note: Kong expect the error to start with "no plugin instance"
            raise PluginServerException("no plugin instance #%s" % iid)

        instance = self.instances[iid]
        instance.reset_expire_ts()
        cls = instance.cls
        phase = event['EventName']

        eid = self.event_id
        self.event_id = eid + 1

        if self.use_multiprocess:
            ch, child_ch = multiprocessing.Pipe(duplex=True)
            self.events[eid] = ch
            self._process_pool.apply_async(
                _handler_event_func,
                (getattr(cls, phase), child_ch, self.lua_style, ),
            )
        elif self.use_gevent:
            # plugin communites to Kong (RPC client) in a reverse way
            ch = gChannel()
            self.events[eid] = ch

            gspawn(_handler_event_func,
                   getattr(cls, phase), ch, self.lua_style,
                   )
        else:  # normal threading mode
            ch = Queue()
            child_ch = Queue()
            ch.get, child_ch.get = child_ch.get, ch.get
            self.events[eid] = ch
            t = threading.Thread(
                target=_handler_event_func,
                args=(getattr(cls, phase), child_ch, self.lua_style, ),
            )
            t.setDaemon(True)
            t.start()

        r = ch.get()
        instance.reset_expire_ts()

        return {
            "Data": r,
            "EventId": eid,
        }

    def _step(self, data, is_error):
        eid = data['EventId']
        if eid not in self.events:
            raise PluginServerException("event id %s not found" % eid)
        dd = None
        if 'Data' in data:
            dd = data['Data']
        ch = self.events[eid]
        if is_error:
            ch.put((
                None, dd
            ))
        else:
            ch.put((
                dd, None
            ))

        ret = ch.get()

        if ret == MSG_RET:
            del self.events[eid]

        return {
            "Data": ret,
            "EventId": eid,
        }

    def step(self, data):
        return self._step(data, False)

    def step_error(self, data):
        return self._step(data, True)


for entity in entities:
    setattr(PluginServer, 'step_' + entity, PluginServer.step)

setattr(PluginServer, 'step_multi_map', PluginServer.step)



# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/response.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class response():


    @staticmethod
    def add_header(name: str, value: Any) -> None:
        """

            Adds a response header with the given value. Unlike
            `kong.response.set_header()`, this function does not remove any existing
            header with the same name. Instead, another header with the same name is
            added to the response. If no header with this name already exists on
            the response, then it is added with the given value, similarly to
            `kong.response.set_header().`

        Phases:
            rewrite, access, header_filter, response, admin_api

        Example:
            kong.response.add_header("Cache-Control", "no-cache")

            kong.response.add_header("Cache-Control", "no-store")

        :parameter name: The header name.
        :type name: str
        :parameter value: The header value.
        :type value: Any

        :return: throws an error on invalid input.

        :rtype: None
        """
        pass

    @staticmethod
    def clear_header(name: str) -> None:
        """

            Removes all occurrences of the specified header in the response sent to
            the client.

        Phases:
            rewrite, access, header_filter, response, admin_api

        Example:
            kong.response.set_header("X-Foo", "foo")

            kong.response.add_header("X-Foo", "bar")

            kong.response.clear_header("X-Foo")

            # from here onwards, no X-Foo headers will exist in the response

        :parameter name: The name of the header to be cleared
        :type name: str

        :return: throws an error on invalid input.

        :rtype: None
        """
        pass

    @staticmethod
    def error(status: number, message: Optional[str], headers: Optional[table]) -> None:
        """

            This function interrupts the current processing and produces an error
            response.
            It is recommended to use this function in conjunction with the `return`
            operator, to better reflect its meaning:
            ```lua
            return kong.response.error(500, "Error", {["Content-Type"] = "text/html"})
            ```
            1. The `status` argument sets the status code of the response that
            is seen by the client. The status code must an error code, that is,
            greater than 399.
            2. The optional `message` argument sets the message describing
            the error, which is written in the body.
            3. The optional `headers` argument can be a table specifying response
            headers to send. If specified, its behavior is similar to
            `kong.response.set_headers()`.
              This method sends the response formatted in JSON, XML, HTML or plaintext.
              The actual format is determined using one of the following options, in
              this order:
              - Manually specified in the `headers` argument using the `Content-Type`
              header.
              - Conforming to the `Accept` header from the request.
              - If there is no setting in the `Content-Type` or `Accept` header, the
              response defaults to JSON format. Also see the `Content-Length`
              header in the produced response for convenience.

        Phases:
            rewrite, access, admin_api, header_filter, (only, if, body`, is, nil)

        Example:
            return kong.response.error(403, "Access Forbidden", {

            ["Content-Type"] = "text/plain",

            ["WWW-Authenticate"] = "Basic"

            })

            # -

            return kong.response.error(403, "Access Forbidden")

            # -

            return kong.response.error(403)

        :parameter status: The status to be used (>399).
        :type status: number
        :parameter message: The error message to be used.
        :type message: str
        :parameter headers: The headers to be used.
        :type headers: table

        :return: throws an error on invalid input.

        :rtype: None
        """
        pass

    @staticmethod
    def exit(status: number, body: Optional[Any], headers: Optional[table]) -> None:
        """

            This function interrupts the current processing and produces a response.
            It is typical to see plugins using it to produce a response before Kong
            has a chance to proxy the request (e.g. an authentication plugin rejecting
            a request, or a caching plugin serving a cached response).
            It is recommended to use this function in conjunction with the `return`
            operator, to better reflect its meaning:
            ```lua
            return kong.response.exit(200, "Success")
            ```
            Calling `kong.response.exit()` interrupts the execution flow of
            plugins in the current phase. Subsequent phases will still be invoked.
            For example, if a plugin calls `kong.response.exit()` in the `access`
            phase, no other plugin is executed in that phase, but the
            `header_filter`, `body_filter`, and `log` phases are still executed,
            along with their plugins. Plugins should be programmed defensively
            against cases when a request is **not** proxied to the Service, but
            instead is produced by Kong itself.
            1. The first argument `status` sets the status code of the response that
            is seen by the client.
               In L4 proxy mode, the `status` code provided is primarily for logging
               and statistical purposes, and is not visible to the client directly.
               In this mode, only the following status codes are supported:
               * 200 - OK
               * 400 - Bad request
               * 403 - Forbidden
               * 500 - Internal server error
               * 502 - Bad gateway
               * 503 - Service unavailable
            2. The second, optional, `body` argument sets the response body. If it is
               a string, no special processing is done, and the body is sent
               as-is.  It is the caller's responsibility to set the appropriate
               `Content-Type` header via the third argument.
               As a convenience, `body` can be specified as a table. In that case,
               the `body` is JSON-encoded and has the `application/json` Content-Type
               header set.
               On gRPC, we cannot send the `body` with this function, so
               it sends `"body"` in the `grpc-message` header instead.
               * If the body is a table, it looks for the `message` field in the body,
               and uses that as a `grpc-message` header.
               * If you specify `application/grpc` in the `Content-Type` header, the
               body is sent without needing the `grpc-message` header.
               In L4 proxy mode, `body` can only be `nil` or a string. Automatic JSON
               encoding is not available. When `body` is provided, depending on the
               value of `status`, the following happens:
               * When `status` is 500, 502 or 503, then `body` is logged in the Kong
               error log file.
               * When the `status` is anything else, `body` is sent back to the L4 client.
            3. The third, optional, `headers` argument can be a table specifying
               response headers to send. If specified, its behavior is similar to
               `kong.response.set_headers()`. This argument is ignored in L4 proxy mode.
            Unless manually specified, this method automatically sets the
            `Content-Length` header in the produced response for convenience.

        Phases:
            preread, rewrite, access, admin_api, header_filter, (only, if, body`, is, nil)

        Example:
            return kong.response.exit(403, "Access Forbidden", {

            ["Content-Type"] = "text/plain",

            ["WWW-Authenticate"] = "Basic"

            })

            # -

            return kong.response.exit(403, [[{"message":"Access Forbidden"}]], {

            ["Content-Type"] = "application/json",

            ["WWW-Authenticate"] = "Basic"

            })

            # -

            return kong.response.exit(403, { message = "Access Forbidden" }, {

            ["WWW-Authenticate"] = "Basic"

            })

            # -

            # In L4 proxy mode

            return kong.response.exit(200, "Success")

        :parameter status: The status to be used.
        :type status: number
        :parameter body: The body to be used.
        :type body: Any
        :parameter headers: The headers to be used.
        :type headers: table

        :return: throws an error on invalid input.

        :rtype: None
        """
        pass

    @staticmethod
    def get_header(name: str) -> str:
        """

            Returns the value of the specified response header, as would be seen by
            the client once received.
            The list of headers returned by this function can consist of both response
            headers from the proxied Service _and_ headers added by Kong (e.g. via
            `kong.response.add_header()`).
            The return value is either a `string`, or can be `nil` if a header with
            `name` is not found in the response. If a header with the same name is
            present multiple times in the request, this function returns the value
            of the first occurrence of this header.

        Phases:
            header_filter, response, body_filter, log, admin_api

        Example:
            # Given a response with the following headers:

            # X-Custom-Header: bla

            # X-Another: foo bar

            # X-Another: baz

            kong.response.get_header("x-custom-header") # "bla"

            kong.response.get_header("X-Another")       # "foo bar"

            kong.response.get_header("X-None")          # nil

        :parameter name: The name of the header.
            Header names are case-insensitive and dashes (`-`) can be written as
            underscores (`_`). For example, the header `X-Custom-Header` can also be
            retrieved as `x_custom_header`.
        :type name: str

        :return: The value of the header.

        :rtype: str
        """
        pass

    @staticmethod
    def get_headers(max_headers: Optional[number]) -> Tuple[table, str]:
        """

            Returns a Lua table holding the response headers. Keys are header names.
            Values are either a string with the header value, or an array of strings
            if a header was sent multiple times. Header names in this table are
            case-insensitive and are normalized to lowercase, and dashes (`-`) can be
            written as underscores (`_`). For example, the header `X-Custom-Header` can
            also be retrieved as `x_custom_header`.
            A response initially has no headers. Headers are added when a plugin
            short-circuits the proxying by producing a header
            (e.g. an authentication plugin rejecting a request), or if the request has
            been proxied, and one of the latter execution phases is currently running.
            Unlike `kong.service.response.get_headers()`, this function returns *all*
            headers as the client would see them upon reception, including headers
            added by Kong itself.
            By default, this function returns up to **100** headers. The optional
            `max_headers` argument can be specified to customize this limit, but must
            be greater than **1** and equal to or less than **1000**.

        Phases:
            header_filter, response, body_filter, log, admin_api

        Example:
            # Given an response from the Service with the following headers:

            # X-Custom-Header: bla

            # X-Another: foo bar

            # X-Another: baz

            headers = kong.response.get_headers()

            headers.x_custom_header # "bla"

            headers.x_another[1]    # "foo bar"

            headers["X-Another"][2] # "baz"

        :parameter max_headers: Limits the number of headers parsed.
        :type max_headers: number

        :return: headers A table representation of the headers in the
            response.

        :rtype: table
        :return: err If more headers than `max_headers` were present,
            returns a string with the error `"truncated"`.

        :rtype: str
        """
        pass

    @staticmethod
    def get_raw_body() -> str:
        """

            Returns the full body when the last chunk has been read.
            Calling this function starts buffering the body in
            an internal request context variable, and sets the current
            chunk (`ngx.arg[1]`) to `nil` when the chunk is not the
            last one. When it reads the last chunk, the function returns the full
            buffered body.

        Phases:
            body_filter`

        Example:
            body = kong.response.get_raw_body()

            if body:

                body = transform(body)

            kong.response.set_raw_body(body)

        :return: body The full body when the last chunk has been read,
            otherwise returns `nil`.

        :rtype: str
        """
        pass

    @staticmethod
    def get_source() -> str:
        """

            This function helps determine where the current response originated
            from. Since Kong is a reverse proxy, it can short-circuit a request and
            produce a response of its own, or the response can come from the proxied
            Service.
            Returns a string with three possible values:
            * `"exit"` is returned when, at some point during the processing of the
              request, there has been a call to `kong.response.exit()`. This happens
              when the request was short-circuited by a plugin or by Kong
              itself (e.g. invalid credentials).
            * `"error"` is returned when an error has happened while processing the
              request. For example, a timeout while connecting to the upstream
              service.
            * `"service"` is returned when the response was originated by successfully
              contacting the proxied Service.

        Phases:
            header_filter, response, body_filter, log, admin_api

        Example:
            if kong.response.get_source() == "service":

                kong.log("The response comes from the Service")

            elseif kong.response.get_source() == "error":

                kong.log("There was an error while processing the request")

            elseif kong.response.get_source() == "exit":

                kong.log("There was an early exit while processing the request")

        :return: The source.

        :rtype: str
        """
        pass

    @staticmethod
    def get_status() -> number:
        """

            Returns the HTTP status code currently set for the downstream response (as
            a Lua number).
            If the request was proxied (as per `kong.response.get_source()`), the
            return value is the response from the Service (identical to
            `kong.service.response.get_status()`).
            If the request was _not_ proxied and the response was produced by Kong
            itself (i.e. via `kong.response.exit()`), the return value is
            returned as-is.

        Phases:
            header_filter, response, body_filter, log, admin_api

        Example:
            kong.response.get_status() # 200

        :return: status The HTTP status code currently set for the
            downstream response.

        :rtype: number
        """
        pass

    @staticmethod
    def set_header(name: str, value: Any) -> None:
        """

            Sets a response header with the given value. This function overrides any
            existing header with the same name.
            Note: Underscores in header names are automatically transformed into dashes
            by default. If you want to deactivate this behavior, set the
            `lua_transform_underscores_in_response_headers` Nginx config option to `off`.
            This setting can be set in the Kong Config file:
                nginx_http_lua_transform_underscores_in_response_headers = off
            Be aware that changing this setting might break any plugins that
            rely on the automatic underscore conversion.

        Phases:
            rewrite, access, header_filter, response, admin_api

        Example:
            kong.response.set_header("X-Foo", "value")

        :parameter name: The name of the header
        :type name: str
        :parameter value: The new value for the header.
        :type value: Any

        :return: throws an error on invalid input.

        :rtype: None
        """
        pass

    @staticmethod
    def set_headers(headers: table) -> None:
        """

            Sets the headers for the response. Unlike `kong.response.set_header()`,
            the `headers` argument must be a table in which each key is a string
            corresponding to a header's name, and each value is a string, or an
            array of strings.
            The resulting headers are produced in lexicographical order. The order of
            entries with the same name (when values are given as an array) is
            retained.
            This function overrides any existing header bearing the same name as those
            specified in the `headers` argument. Other headers remain unchanged.

        Phases:
            rewrite, access, header_filter, response, admin_api

        Example:
            kong.response.set_headers({

            ["Bla"] = "boo",

            ["X-Foo"] = "foo3",

            ["Cache-Control"] = { "no-store", "no-cache" }

            })

            # Will add the following headers to the response, in this order:

            # X-Bar: bar1

            # Bla: boo

            # Cache-Control: no-store

            # Cache-Control: no-cache

            # X-Foo: foo3

        :parameter headers: 
        :type headers: table

        :return: throws an error on invalid input.

        :rtype: None
        """
        pass

    @staticmethod
    def set_raw_body(body: str) -> None:
        """

            Sets the body of the response.
            The `body` argument must be a string and is not processed in any way.
            This function can't change the `Content-Length` header if one was
            added. If you decide to use this function, the `Content-Length` header
            should also be cleared, for example in the `header_filter` phase.

        Phases:
            body_filter`

        Example:
            kong.response.set_raw_body("Hello, world!")

            # or

            body = kong.response.get_raw_body()

            if body:

                body = transform(body)

            kong.response.set_raw_body(body)

        :parameter body: The raw body.
        :type body: str

        :return: throws an error on invalid inputs.

        :rtype: None
        """
        pass

    @staticmethod
    def set_status(status: number) -> None:
        """

            Allows changing the downstream response HTTP status code before sending it
            to the client.

        Phases:
            rewrite, access, header_filter, response, admin_api

        Example:
            kong.response.set_status(404)

        :parameter status: The new status.
        :type status: number

        :return: throws an error on invalid input.

        :rtype: None
        """
        pass

    pass


# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/vault.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class vault():


    @staticmethod
    def get(reference: str) -> Tuple[str, str]:
        """

            Resolves the passed in reference and returns the value of it.

        Example:
            value, err = kong.vault.get("{vault://env/cert/key}")

        :parameter reference: reference to resolve
        :type reference: str

        :return: resolved value of the reference

        :rtype: str
        :return: error message on failure, otherwise `nil`

        :rtype: str
        """
        pass

    @staticmethod
    def is_reference(reference: str) -> bool:
        """

            Checks if the passed in reference looks like a reference.
            Valid references start with '{vault://' and end with '}'.
            If you need more thorough validation,
            use `kong.vault.parse_reference`.

        Example:
            kong.vault.is_reference("{vault://env/key}") # true

            kong.vault.is_reference("not a reference")   # false

        :parameter reference: reference to check
        :type reference: str

        :return: `true` is the passed in reference looks like a reference, otherwise `false`

        :rtype: bool
        """
        pass

    @staticmethod
    def parse_reference(reference: str) -> Tuple[table, str]:
        """

            Parses and decodes the passed in reference and returns a table
            containing its components.
            Given a following resource:
            ```lua
            "{vault://env/cert/key?prefix=SSL_#1}"
            ```
            This function will return following table:
            ```lua
            {
              name     = "env",  -- name of the Vault entity or Vault strategy
              resource = "cert", -- resource where secret is stored
              key      = "key",  -- key to lookup if the resource is secret object
              config   = {       -- if there are any config options specified
                prefix = "SSL_"
              },
              version  = 1       -- if the version is specified
            }
            ```

        Example:
            ref, err = kong.vault.parse_reference("{vault://env/cert/key?prefix=SSL_#1}") # table

        :parameter reference: reference to parse
        :type reference: str

        :return: a table containing each component of the reference, or `nil` on error

        :rtype: table
        :return: error message on failure, otherwise `nil`

        :rtype: str
        """
        pass

    pass


#!/usr/bin/env python3
import os
import imp
from setuptools import setup

PROJ_NAME = 'kong-pdk'
PACKAGE_NAME = 'kong_pdk'

PROJ_METADATA = '%s.json' % PROJ_NAME

here = os.path.abspath(os.path.dirname(__file__))

try:
    README = open(os.path.join(here, 'README.md')).read()
except:  # noqa: E722 do not use bare 'except
    README = ""
try:
    CHANGELOG = open(os.path.join(here, 'CHANGELOG.md')).read()
except:  # noqa: E722 do not use bare 'except
    CHANGELOG = ""
VERSION = "%.2f" % imp.load_source('version', os.path.join(here, '%s/const.py' % PACKAGE_NAME)).__version__

packages = [
    'kong_pdk',
    'kong_pdk/pdk',
    'kong_pdk/pdk/kong',
    'kong_pdk/pdk/kong/client',
    'kong_pdk/pdk/kong/ctx',
    'kong_pdk/pdk/kong/nginx',
    'kong_pdk/pdk/kong/service',
]
requires = ['gevent', 'msgpack']

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description='Kong PDK for Python and Plugin Server',
    long_description=README + '\n\n' + CHANGELOG,
    long_description_content_type="text/markdown",
    author='Kong',
    url='https://github.com/Kong/kong-python-pdk',
    packages=packages,
    package_dir={'requests': 'requests'},
    include_package_data=True,
    install_requires=requires,
    license="Apache-2.0",
    license_files=('LICENSE',),
    zip_safe=False,
    classifiers=(
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython'
    ),
    requires=requires,
    entry_points={'console_scripts': ["kong-python-pluginserver = kong_pdk.cli:start_server"]},
    package_data={"kong_pdk": ["*.pyi", "**/*.pyi", "**/**/*.pyi", "**/**/**/*.pyi"]}
)



# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/ctx.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str

from .shared import shared as cls_shared

class ctx():

    shared = cls_shared

    pass


# AUTO GENERATED BASED ON Kong 2.7.x, DO NOT EDIT
# Original source path: kong/pdk/cluster.lua

from typing import TypeVar, Any, Union, List, Mapping, Tuple, Optional

number = TypeVar('number', int, float)
table = TypeVar('table', List[Any], Mapping[str, Any])
# XXX
cdata = Any
err = str


class cluster():


    @staticmethod
    def get_id() -> Tuple[str, str]:
        """

            Returns the unique ID for this Kong cluster. If Kong
            is running in DB-less mode without a cluster ID explicitly defined,
            then this method returns `nil`.
            For hybrid mode, all control planes and data planes belonging to the same
            cluster return the same cluster ID. For traditional database-based
            deployments, all Kong nodes pointing to the same database also return
            the same cluster ID.

        Example:
            id, err = kong.cluster.get_id()

            if err:

                # handle errorif not id:

                # no cluster ID is available# use id here

        :return: The v4 UUID used by this cluster as its ID.

        :rtype: str
        :return: An error message.

        :rtype: str
        """
        pass

    pass